;;; -*- Lisp -*-
;;;
;;; continuation of cmn1.lisp

#+(or KCL (and Excl (not cltl2))) (in-package :cmn :use '(:lisp :pcl :loop))
#+(or CCL Lucid (and excl cltl2)) (in-package :cmn)

;;;
;;; ----------------    notes (leger lines, dots)
;;;

(defclass rhythm-mixin ()
  ((dots :initarg :dots :initform 0 :reader dots)
   (flags :initarg :flags :initform 0 :reader flags)
   (quarters :initarg :quarters :initform nil :reader quarters)
   (ratified :initarg :ratified :initform nil :reader ratified)))

(defclass rhythm (rhythm-mixin)
  ((dots :accessor dots)
   (flags :accessor flags)
   (quarters :accessor quarters)
   (ratified :accessor ratified)))

(defclass write-protected-rhythm (write-protect rhythm-mixin) ())

(defmethod rhythm-p ((obj t)) nil)
(defmethod rhythm-p ((obj rhythm-mixin))
  (and (not (note-p obj))
       (not (chord-p obj))
       (not (rest-p obj))))

(defmethod rcopy (rhy obj)
  (setf (dots obj) (dots rhy))
  (setf (quarters obj) (fratify (quarters rhy)))
  (setf (ratified obj) (ratified rhy))
  (setf (flags obj) (flags rhy)))

(defmethod copy ((rhythm rhythm-mixin) &optional object)
  (let ((new-rhythm (if (not object) (make-rhythm)
		      (if (write-protected object) (copy object)
			object))))
    (setf (dots new-rhythm) (dots rhythm))
    (if (quarters rhythm)
	(setf (quarters new-rhythm) 
	  (fratify (quarters rhythm))))
    (setf (ratified new-rhythm) (ratified rhythm))
    (setf (flags new-rhythm) (flags rhythm))
    (if (next-method-p) (call-next-method rhythm new-rhythm))
    new-rhythm))

(defmethod descry ((rhythm rhythm-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A"
	  (if (not controller) "(rhythm" "")
	  (if (and (quarters rhythm) (not (zerop (quarters rhythm)))) (format nil " :quarters ~A" (quarters rhythm)) "")
	  (if (and (dots rhythm) (not (zerop (dots rhythm)))) (format nil " :dots ~A" (dots rhythm)) "")
	  (if (and (flags rhythm) (not (zerop (flags rhythm)))) (format nil " :flags ~A" (flags rhythm)) "")
	  (if (ratified rhythm) " :ratified t" "")
	  (if (next-method-p) (call-next-method rhythm stream (or controller rhythm)) "")
	  (if (not controller) ")" "")))

(defmethod identify ((rhythm rhythm-mixin))
  (format nil "(rhythm ~A)" (rhythm-name (quarters rhythm))))

(defmethod notify ((rhy rhythm-mixin) &optional objects) 
  (declare (ignore objects))
  (cmn-error "~A seems to be unattached to anything" (identify rhy)))
;;; rhythm-mixin needed for notes, chords, rests -- never independent

(defmacro define-rhythm (name dots flags quarters)
  `(defvar ,name (make-instance 'write-protected-rhythm :dots ,dots :flags ,flags :quarters ,quarters)))

(define-rhythm double-whole      0 0 8)
(define-rhythm whole             0 0 4)
(define-rhythm half              0 0 2)
(define-rhythm quarter           0 0 1)
(define-rhythm eighth            0 1 1/2)
(define-rhythm sixteenth         0 2 1/4)
(define-rhythm thirty-second     0 3 1/8)
(define-rhythm sixty-fourth      0 4 1/16)
(define-rhythm one-twenty-eighth 0 5 1/32)

(define-rhythm q     0 0 1)
(define-rhythm q.    1 0 3/2)
(define-rhythm q..   2 0 7/4)
(define-rhythm h     0 0 2)
(define-rhythm h.    1 0 3)
(define-rhythm h..   2 0 7/2)
(define-rhythm w     0 0 4)
(define-rhythm w.    1 0 6)
(define-rhythm dw    0 0 8)
(define-rhythm e     0 1 1/2)
(define-rhythm e.    1 1 3/4)
(define-rhythm e..   2 1 7/8)
(define-rhythm s     0 2 1/4)
(define-rhythm s.    1 2 3/8)
(define-rhythm s..   2 2 7/16)

(define-rhythm ts     0 2 1/6)
(define-rhythm te     0 1 1/3)
(define-rhythm tq     0 0 2/3)
(define-rhythm th     0 0 4/3)

(defun rhythm-name (qrhy)
  (let ((quarters qrhy))
    (if (= quarters 1) "q"
      (if (= quarters 3/2) "q."
	(if (= quarters 7/4) "q.."
	  (if (= quarters 2) "h"
	    (if (= quarters 3) "h."
	      (if (= quarters 7/2) "h.."
		(if (= quarters 4) "w"
		  (if (= quarters 6) "w."
		    (if (= quarters 8) "dw"
		      (if (= quarters 1/2) "e"
			(if (= quarters 3/4) "e."
			  (if (= quarters 7/8) "e.."
			    (if (= quarters 1/4) "s"
			      (if (= quarters 3/8) "s."
				(if (= quarters 7/16) "s.."
				  (if (= quarters 1/6) "ts"
				    (if (= quarters 1/3) "te"
				      (if (= quarters 2/3) "tq"
					(if (= quarters 4/3) "th"
					  (if (not quarters) "none"
					    (format nil "(rq ~A)" (fratify quarters))))))))))))))))))))))))

(defun how-many-flags (num &optional (cur-flags 0))
  (if (not (plusp num)) (cmn-error "impossible duration: ~1,3F" num))
  (if (or (>= num 1.0)
	  (and (> num .5) 
	       (or (> cur-flags 1) 
		   (and (/= num 3/4) (/= num 7/8)))))
      cur-flags
    (how-many-flags (* 2 num) (1+ cur-flags))))

(defun how-many-dots (rat)
  (let ((num (if (ratiop rat) 
		 (numerator rat)
	       (first (ratify rat)))))
    (if (= num 3) 1 (if (= num 7) 2 0))))

(defun rq (rat)
  (make-rhythm 
   :dots (how-many-dots rat)
   :flags (how-many-flags rat)
   :quarters rat))

(defun whole-note-p (note)
  (> (quarters note) 3.5))



;;; methods for audible class (in cmn1.lisp to make loader happier)

(defmethod descry ((note audible-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A~A~A~A~A~A~A~A"
	  (if (slurs note)  (format nil " :slurs (list~{ ~A~})" (loop for slur in (slurs note) collect (descry slur))) "")
	  (if (beams note) (format nil " :beams ~A~%                    " (descry (beams note))) "")
	  (if (beamed note) (format nil " :beamed ~A" (beamed note)) "")
	  (if (octaved note) " :octaved t" "")
	  (if (outer-beam note)
	      (format nil " :outer-beam '(~{~1,3F ~})" (outer-beam note))
	    "")
	  (if (get-tremolo note) (format nil " :tremolo ~A" (descry (get-tremolo note))) "")
	  (if (ties note) (format nil " :ties (list~{ ~A~})" (loop for tie in (ties note) collect (descry tie))) "")
	  (if (stem-direction note) (format nil " :stem-direction :~(~A~)" (stem-direction note)) "")
	  (if (tie-direction note) (format nil " :tie-direction :~(~A~)" (tie-direction note)) "")
	  (if (slur-direction note) (format nil " :slur-direction :~(~A~)" (slur-direction note)) "")
	  (if (stem-end note) (format nil " :stem-end ~1,3F" (stem-end note)) "")
	  (if (collision note) (format nil " :collision ~A" (collision note)) "")
	  (if (store-data note) (format nil " :store-data ~A" (store-data note)) "")
	  (if (next-method-p) (call-next-method note stream (or controller note)) "")))

(defmethod copy ((note audible-mixin) &optional new-chord)
  (when (slurs note) (copy-slurs note new-chord))
  (if (beams note) (setf (beams new-chord) (copy (beams note))))
  (setf (beamed new-chord) (beamed note))
  (setf (octaved new-chord) (octaved note))
  (setf (outer-beam new-chord) (outer-beam note))
  (if (get-tremolo note) (setf (get-tremolo new-chord) (copy (get-tremolo note))))
  (if (ties note) (setf (ties new-chord) (copy-ties new-chord note)))
  (setf (audible-stem-direction new-chord) (audible-stem-direction note))
  (setf (audible-tie-direction new-chord) (audible-tie-direction note))
  (setf (audible-slur-direction new-chord) (audible-slur-direction note))
  (setf (stem-end new-chord) (stem-end note))
  (setf (audible-collision new-chord) (audible-collision note))
  (if (next-method-p) (call-next-method note new-chord))
  new-chord)





(defclass note-mixin (audible-mixin rhythm-mixin staff-relative-mixin score-object-mixin)
  ((line :initarg :line :initform nil :reader line :reader note-line)
   (note-head :initarg :note-head :initform nil :reader note-head) ; a type like :diamond, nil = choose what seems best from context
   (sign :initarg :sign :initform nil :reader sign :reader note-sign)
   (cclass :initarg :cclass :initform nil :reader cclass)
   (pitch :initarg :pitch :initform nil :reader pitch)
   (octave :initarg :octave :initform nil :reader octave)
   (head-quarters :initform nil :reader head-quarters)
   (stem-dy :initform nil :reader stem-dy)
   (stem-mark :initform nil :reader stem-mark)))

(defclass note (note-mixin audible rhythm staff-relative-mixin score-object) 
  ((line :accessor line :accessor note-line)
   (note-head :accessor note-head :initform nil)
   (sign :accessor sign :accessor note-sign)
   (cclass :accessor cclass)
   (pitch :accessor pitch)
   (octave :accessor octave)
   (head-quarters :accessor head-quarters)
   (stem-dy :accessor stem-dy)
   (stem-mark :accessor stem-mark)
   (quarters :initform 1)))

(self-action stem-direction setf-stem-direction)
(self-action sign setf-sign)
(self-action pitch setf-pitch)
(self-action octave setf-octave)
(self-action dots setf-dots)
(self-action flags setf-flags)
(self-action note-head setf-note-head)
(self-action beams setf-beams)
;(self-action cclass setf-cclass)	;for add-note internally
(self-action ratified setf-ratified)
(self-action quarters setf-quarters)
(self-action beamed setf-beamed)
(self-action stem-dy setf-stem-dy)
(self-action stem-mark setf-stem-mark)

(defmethod add-to-marks ((note note) objects) (setf (marks note) (append (marks note) objects)))
(defmethod add-to-slurs ((note note) object) (push object (slurs note)))
(defmethod add-to-ties ((note note) object) (push object (ties note))) ; internal -- in automatic tie creation

(defparameter old-octave 4)

(defmethod setf-rhythm ((note note) ur-val)
  (if (minusp ur-val) (cmn-error "attempt to set negative rhythm"))
  (let ((val (fratify ur-val)))
    (setf (odb-duration note) val)
    (setf (dots note) (how-many-dots val))
    (setf (flags note) (how-many-flags val))
    (setf (quarters note) val)))

(defun rhythm (dur) (make-self-acting :action #'setf-rhythm :arguments dur))

(defun stem-up () (stem-direction :up))
(defun stem-down () (stem-direction :down))
(defun no-stem () (stem-direction :none))
(defvar stem-up (make-self-acting :action #'setf-stem-direction :arguments :up))
(defvar stem-down (make-self-acting :action #'setf-stem-direction :arguments :down))
(defvar no-stem (make-self-acting :action #'setf-stem-direction :arguments :none))

(defun stem-eq (s1 s2)
  (or (eq s1 s2)
      (and (member s1 '(:up :up?)) (member s2 '(:up :up?)))
      (and (member s1 '(:down :down?)) (member s2 '(:down :down?)))))

(defclass write-protected-note (write-protect note-mixin) ())

(defmethod note-p ((obj t)) nil)
(defmethod note-p ((obj note-mixin)) t)

(defmethod descry ((note note-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A~A~A~A~A"
	  (if (not controller) "(note" "")
	  (if (note-head note) 
	      (format nil " :note-head :~(~A~)~%~A" 
		      (note-head note)
		      prewhitespace) 
	    "")
	  (if (sign note) 
	      (format nil " :sign ~A~%~A" 
		      (descry (sign note))
		      prewhitespace) 
	    "")
	  (if (cclass note) (format nil " :cclass ~A" (cclass note)) "")
	  (if (pitch note) (format nil " :pitch ~A" (pitch note)) "")
	  (if (octave note) (format nil " :octave ~A" (octave note)) "")
	  (if (note-line note) (format nil " :line ~A" (note-line note)) "")
	  (if (head-quarters note) (format nil " :head-quarters ~A" (head-quarters note)) "")
	  (if (stem-dy note) (format nil " :stem-dy ~A" (stem-dy note)) "")
	  (if (next-method-p) (call-next-method note stream (or controller note)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((note note-mixin) &optional object)
  (let ((new-note (if (not object) (make-simple-note)
		    (if (write-protected object) (copy object)
		      object))))
    (if (next-method-p) (call-next-method note new-note))
    (if (note-head note) (setf (note-head new-note) (note-head note)))
    (setf (note-sign new-note) (note-sign note))
    (setf (note-line new-note) (note-line note))
    (setf (cclass new-note) (cclass note))
    (setf (pitch new-note) (pitch note))
    (setf (head-quarters new-note) (head-quarters note))
    (setf (stem-dy new-note) (stem-dy note))
    (setf (stem-mark new-note) (stem-mark note))
    (if (octave note)
	(progn
	  (setf (octave new-note) (octave note))
	  (setf old-octave (octave note)))
      (setf (octave new-note) old-octave))
    new-note))

(defmethod short-note ((note note-mixin) &key x0)
  (make-note :x0 (or x0 (box-x0 note)) :x1 (box-x1 note) :y0 (box-y0 note) :y1 (box-y1 note)
	     :staff-y0 (%staff-y0 note) :stem-direction (audible-stem-direction note)
	     :tie-direction (audible-tie-direction note) :slur-direction (audible-slur-direction note)
	     :center (center note) :ties (if (ties note) t nil) :quarters (quarters note)
	     :stem-end (stem-end note) :line (note-line note) :pitch (pitch note) :octave (octave note)))

(defun note-name (note) 
  (let* ((sign (note-sign note))
	 (nat-sign (and sign (marks sign) (find-if #'(lambda (n) 
						       (and (sundry-p n) 
							    (member (sundry-name n) (list :natural-sharp :natural-flat))))
						   (marks sign))))
	 (nat-sign-name (and nat-sign (sundry-name nat-sign)))
	 (funny-sign (and sign (not nat-sign) (not (member sign (list sharp flat natural)))))
	 (name (case (pitch note)
		 (0 (if (eq sign sharp) "bs" (if (eq sign natural) "cn" "c")))
		 (1 (if (eq sign sharp) "cs" "df"))
		 (2 (if (eq sign natural) "dn" "d"))
		 (3 (if (eq sign sharp) "ds" "ef"))
		 (4 (if (eq sign natural) "en" "e"))
		 (5 (if (eq sign natural) "fn" (if (eq sign sharp) "es" "f")))
		 (6 (if (eq sign sharp) "fs" "gf"))
		 (7 (if (eq sign natural) "gn" "g"))
		 (8 (if (eq sign sharp) "gs" "af"))
		 (9 (if (eq sign natural) "an" "a"))
		 (10 (if (eq sign sharp) "as" "bf"))
		 (11 (if (eq sign natural) "bn" (if (eq sign flat) "cf" "b")))
		 (12 (if (eq sign sharp) "bs" "c")))))
    (format nil "~A~D~A~(~A~)" name (octave note) 
	    (if (or nat-sign funny-sign) " " "") 
	    (if funny-sign (sign-name sign)
	      (if nat-sign nat-sign-name "")))))

(defmethod brief-identify ((note note-mixin))
  (format nil "(~A ~A~A)"
	  (note-name note) 
	  (if (not (halve-duration note)) 
	      (rhythm-name (quarters note))
	    (rhythm-name (* 2 (quarters note))))
	  (if (onset note) (format nil " (onset ~A)" (fratify (onset note))) "")))

(defmethod brief-identify ((whatever t))
  whatever)

(defmethod identify ((note note-mixin))
  (format nil "(~A ~A~A~A~A~A~A~A~A~A~A~A)" 
	  (note-name note) 
	  (if (not (halve-duration note)) 
	      (rhythm-name (quarters note))
	    (rhythm-name (* 2 (quarters note))))
	  (if (onset note) (format nil " (onset ~A)" (fratify (onset note))) "")
	  (if (and (note-head note)
		   (not (eq (note-head note) :normal)))
	      (format nil " (note-head :~(~A~))" (note-head note))
	    "")
	  (if (eq (audible-stem-direction note) :up) " stem-up"
	    (if (eq (audible-stem-direction note) :down) " stem-down"
	      (if (eq (audible-stem-direction note) :none) " no-stem"
		"")))
	  (if (stem-dy note) (format nil " (stem-dy ~A)" (stem-dy note)) "")
	  (if (audible-tie-direction note) (format nil " (tie-direction :~(~A~))" (audible-tie-direction note)) "")
	  (if (audible-slur-direction note) (format nil " (slur-direction :~(~A~))" (audible-slur-direction note)) "")
	  (if (store-data note) (identify-store-data note) "")
	  (if (ties note) 
	      (format nil "~{ ~(~A~)~}" (sorted-ties (ties note)))
	    "")
	  (if (get-tremolo note) 
	      (identify (get-tremolo note))
	    "")
	  (the-usual-suspects note)))

(defvar with-errors t)

(defun ur-note (new-note &rest ur-objects)
  (setf *cmn-owning-object* new-note)
  (let ((objects ur-objects))
    (loop while objects do
      (let ((object (pop objects)))
	(setf *cmn-object* object)
	(when object
	  (if (self-acting-p object)
	      (funcall (action object) new-note (arguments object))
	    (if (rhythm-p object)
		(rcopy object new-note)
	      (if (and (score-object-list-p object) 
		       (not (glyph-list-p object)))
		  (setf objects (append (disgorge object) objects))
		(if (score-object-p object)
		    (if (accidental-p object)
			(setf (note-sign new-note) object)
		      (if (or (sundry-p object) (pause-p object) (dynamics-p object) (glyph-list-p object))
			  (add-to-marks new-note (list (if (write-protected object) (copy object) object)))
			(copy object new-note)))
		  (if (text-p object)
		      (add-to-marks new-note (list object))
		    (if with-errors
			(cmn-warn "odd argument to note: ~A to ~A" object new-note)))))))))))
  (when (halve-duration new-note)
    (setf (head-quarters new-note) (quarters new-note))
    (if (quarters new-note) (setf (quarters new-note) (* .5 (quarters new-note))))
    (if (odb-duration new-note) (setf (odb-duration new-note) (* .5 (odb-duration new-note)))))
  (setf (octaved new-note) (or (octaved new-note) octave-stack))
  (cmn-tick-pipe new-note)
  new-note)

(defun note (&rest objects)
  (apply #'ur-note (make-simple-note) objects))

(defmethod notify ((note note) &optional objects)
  (if (listp objects)
      (apply #'ur-note note objects)
    (ur-note note objects)))

(defun get-note-head (head-quarters &optional (head-type nil))
  (if (or (not head-type) 
	  (member head-type '(:normal :small)))
      (if (= head-quarters 8)
	  %dblwholenote
	(if (>= head-quarters 4)
	    %wholenote
	  (if (or (>= head-quarters 2) (= head-quarters 4/3))
	      %halfnotehead
	    %quarternotehead)))
    (if (or (sundry-p head-type)
	    (glyph-p head-type))
	head-type
      (case head-type
	(:diamond (if (< head-quarters 2) %diamondfill %diamond))
	(:tremolo %halfnotehead)
	(:artificial-harmonic %diamond)
	(:triangle (if (< head-quarters 2) 
		       %trianglefill 
		     (make-sundry :name :triangle 
				  :mark #'(lambda (mark note score &rest rest)
					    (declare (ignore mark note rest))
					    (setf (line-width score) .04)
					    (rmoveto score 0 -.125)
					    (rlineto score .15 .25)
					    (rlineto score .15 -.25)
					    (rlineto score -.3 0)
					    (draw score)
					    (setf (line-width score) 0)))))
	(:square (if (< head-quarters 2) %squarefill %square))
	(:x %rhythmX)
	(:none %space)
	(otherwise (cmn-error "unknown note head type: ~A" head-type))))))

(defmethod rx ((note note-mixin))
  (g-rx (get-note-head (or (head-quarters note) (quarters note)) (note-head note))))

(defmethod g-rx ((sundry sundry-mixin)) .3)

(defun draw-flags (score nflags stem-up stem-end x0)
  (let* ((y0 (+ stem-end (if stem-up -.1 .1)))
	 (incr (if stem-up (- (flag-dy score)) (flag-dy score))))
    #+Petrucci-Medium (if (and stem-up (curvy-flags score)) (incf x0 .26))
    (moveto score x0 (+ y0 
			(* incr nflags) 
			(if stem-up
			    (flag-up-y-offset score) 
			  (flag-down-y-offset score))))
    #+Petrucci-Medium (if (and (curvy-flags score) (not stem-up)) (decf y0 .1))
    (if (curvy-flags score)
	(progn
	  (if stem-up
	      (simple-show score %eighthflagup) 
	    (simple-show score %eighthflagdown))
#-KCL	  (loop for i from 1 below nflags and y from y0 by incr do
	    (moveto score x0 y)
	    (if stem-up
		(simple-show score %extendflagup) 
	      (simple-show score %extendflagdown)))
#+KCL     (do ((i 1 (1+ i))
	       (y y0 (+ y incr)))
	      ((>= i nflags))
	    (moveto score x0 y)
	    (if stem-up
		(simple-show score %extendflagup) 
	      (simple-show score %extendflagdown)))
         )	    
      (let* ((bx0 (+ x0 (if stem-up (- .3 (half-stem-width score)) (half-stem-width score))))
	     (by0 (if stem-up (- stem-end .1) stem-end)))
	#-KCL	(loop for i from 0 below nflags and y from by0 by incr do
		  (fill-in score :rectangle t :path (list bx0 y .3 .1)))
	
	#+KCL	(do ((i 0 (1+ i))
		     (y by0 (+ y incr)))
		    ((>= i nflags))
		  (fill-in score :rectangle t :path (list bx0 y .3 .1)))
        ))))



(defun draw-stem (score stem-up x0 y0 y1 note-head-width &optional smark)
  (let* ((stem-width (stem-width score))
	 (half-stem-width (half-stem-width score))
	 (xs (+ x0 (if stem-up (- note-head-width half-stem-width) half-stem-width))))
    (if smark (funcall smark score xs y0 y1))
    (optimize-stem score xs y0 y1 stem-width)))

(defun get-stem-end (score note stem-up &optional the-rest)
  (let* ((mid-staff (+ (or (%staff-y0 note) 0) 
		       (if (member :grace-note the-rest)
			   (if stem-up
			       (/ .125 (grace-note-size score))
			     (/ .875 (grace-note-size score))) 
			 (half-staff-dy score))))
	 (nflags (or (flags note) 0))
	 (dots (or (dots note) 0))
	 (note-y0 (+ (box-y0 note) (dy note)))
	 (note-line (if stem-up (minimum-line note) (maximum-line note)))
	 (dot-room (if (and stem-up
			    (plusp dots)
			    (plusp nflags))
		       (flag-to-dot-distance score)
		     0))
	 (tremolo-room (if (get-tremolo note)
			   (* (or (tremolo-slashes (get-tremolo note)) 0) (staff-line-separation score))
			 0))
	 (flag-room (+ (* nflags 2 (staff-line-separation score))
		       (if (and (not stem-up)
				(plusp nflags)
				(> (note-head-size score) 1.1))
			   (* .5 (- (note-head-size score) 1.1))
			 0))))
    
    (if (or (<= -1 note-line 9)		; not leger lines, so just make stem long enough
	    (and stem-up (> note-line 9))
	    (and (not stem-up) (< note-line -1)))
	(if stem-up
	    (+ note-y0 (max .3 (+ flag-room tremolo-room)) dot-room .5)
	  (- note-y0 (max .3 (+ flag-room tremolo-room)) .5))
      (if (zerop nflags)
	  mid-staff
	(let* ((stem-length (+ (flagged-stem-length score) flag-room tremolo-room dot-room))
	       (cur-stem (abs (- note-y0 mid-staff)))
	       (extra-stem (max 0 (- stem-length cur-stem))))
	  (if stem-up
	      (incf mid-staff extra-stem)
	    (decf mid-staff extra-stem))
	  mid-staff)))))

(defun draw-tremolo (score note slashes paranoid x0 y0 y1 stem-dir &optional (y-offset 0.0))
  ;; tremolo is the number of tremolo slashes desired.  We assume on the stem here, or on the entire note.
  ;; "note" may be a chord.
  ;; x0 y0 y1 describe the stem position
  ;; if no stem (as in whole note with tremolo), y1 is nil but stem-dir tells us where to put the slashes
  (when (plusp slashes)
    (let* ((stem-up (member stem-dir '(:up :up?)))
	   (thickness (tremolo-thickness score))
	   (ur-incr (- (* 2 thickness) .01))
	   (incr (if stem-up ur-incr (- ur-incr)))
	   (slant (tremolo-slant score))
	   (width (tremolo-width score))
	   (x-off (- x0 (- (* .5 width) .01)))
	   (bottom (+ (if stem-up
			  (+ y0 .2 y-offset)
			(- y0 (+ .2 thickness slant y-offset)))
		      (if (= slashes 1) incr 0))))
      
      #-KCL (loop for i from 0 below slashes and y-off from bottom by incr do
	      (moveto score x-off y-off)
	      (lineto score x-off (+ y-off thickness))
	      (lineto score (+ x-off width) (+ y-off thickness slant))
	      (lineto score (+ x-off width) (+ y-off slant))
	      (fill-in score :closepath t))
      
      #+KCL (do ((i 0 (1+ i))
		 (y-off bottom (+ y-off incr)))
		((>= i slashes))
	      (moveto score x-off y-off)
	      (lineto score x-off (+ y-off thickness))
	      (lineto score (+ x-off width) (+ y-off thickness slant))
	      (lineto score (+ x-off width) (+ y-off slant))
	      (fill-in score :closepath t))
      
      (when paranoid
	(moveto score 
		(- (if stem-up x-off x0) .25)
		(+ .125 (max (+ (%staff-y0 note) 1.0) y0 (or y1 0))))
	(show score (%%text :letters "trem." :font-name (tremolo-font score) :font-scaler (tremolo-font-scaler score)))))))

(defun draw-two-note-tremolo (score note tremolo)
  ;; make three "beams" between note and (tremolo-note tremolo)
  ;; set stem-direction of next note too
  (if (not (measured tremolo))
      (let* ((next-note (tremolo-note tremolo))
	     (stem-dir (audible-stem-direction note))
	     (above (member stem-dir '(:up :up?)))
	     (x0 (if (or above (>= (head-quarters note) 4)) (box-x1 note) (box-x0 note)))
	     (y0 (+ (%staff-y0 note) (* (line note) (staff-line-separation score))))
	     (y1 (+ (%staff-y0 next-note) (* (line next-note) (staff-line-separation score))))
	     (y-diff (- y1 y0))
	     (x1 (if (and above (< y-diff .5)) (box-x1 next-note) (box-x0 next-note)))
	     (y-incr (min .2 (max -.2 (* .5 y-diff))))
	     (py0 (+ y0 (* .25 y-diff) (if above .25 -.4)))
	     (py1 (+ py0 y-incr))
	     (x-incr (if (>= (head-quarters note) 4) .1 0))
	     (bx0 (+ x0 x-incr (if above -.05 .125)))
	     (bx1 (- x1 x-incr (if above .2 .1))))
	(when (> (- bx1 bx0) (tremolo-maximum-length score))
	  (let ((excess (- bx1 bx0 (tremolo-maximum-length score))))
	    (incf bx0 (* .5 excess))
	    (decf bx1 (* .5 excess))))
	(if (not (stem-is-locked next-note)) (setf (audible-stem-direction next-note) stem-dir))
	(draw-beams score bx0 py0 bx1 py1 3 stem-dir .75))))

(defun get-note-sign (note)
  (if (eq (note-head note) :small)
      (if (eq (note-sign note) sharp)
	  (setf (note-sign note) %smallsharp)
	(if (eq (note-sign note) flat)
	    (setf (note-sign note) %smallflat)
	  (if (eq (note-sign note) natural)
	      (setf (note-sign note) %smallnatural)
	    (note-sign note))))
    (note-sign note)))

(defun note-head-y0-offset (note stem-up)
  ;; note and chord
  (let ((head (note-head note)))
    (if (or (not head)
	    (eq head :normal))
	(if (>= 3.5 (quarters note) 2)
	    (if stem-up .02 -.02)
	  0)
      (if (eq head :triangle)
	  -.1
	(if (eq head :x)
	    (if stem-up .12 -.12)
	  0)))))

(defun display-note-slurs-and-brackets (object score &rest rest)
  (if (marks object) 
      (loop for mark in (marks object) do 
	(if (and (sundry-p mark)
		 (member (sundry-name mark) '(:slur :beat)))
	    (apply #'display mark object score rest)))))

(defmethod display ((note note-mixin) chord score &rest rest)
  (let* ((line (note-line note))
	 (nx0 0)
	 (nflags (or (flags note) 0))
	 (dots (or (dots note) 0))
	 (nh (get-note-head (or (head-quarters note) (quarters note)) (note-head note)))
	 (nx1 0)
	 (nsx (* .3 (- 1.0 (note-head-size score))))
	 (sls (staff-line-separation score))
	 (x0-note (box-x0 note))
	 (leger-len (leger-line-length score))
	 (dx-note (if (member :just-fooling rest) 0 (dxy-dx note)))
	 (dy-note (if (member :just-fooling rest) 0 (dxy-dy note))))
    
    (setf (box-y0 note) (+ (%staff-y0 note) (* line sls)))

    (when (and (not (member :just-fooling rest))
	       (not (member :just-do-it rest)))
      (if (beams note) (display (beams note) note score))
      (if (ties note) (loop for tie in (ties note) do (display tie note score))))
    (if (marks note) (apply #'display-marks note score rest))
    
    (moveto score (+ x0-note dx-note) (+ (box-y0 note) dy-note))
    (if (invisible-matrix-p note)
	(if (marks note) (apply #'display-note-slurs-and-brackets note score rest))
      (progn
	(when (note-sign note)
	  (if (plusp (audible-collision note)) (rmoveto score (- (+ (audible-collision note) .125)) 0))
	  (if (/= nsx 0.0) (rmoveto score nsx 0))
	  (let ((nsgn (get-note-sign note)))
	    (if (not (accidental-p nsgn))
		(show score nsgn)
	      (display nsgn note score)))
	  (if (plusp (audible-collision note))
	      (moveto score (+ x0-note dx-note) (+ (box-y0 note) dy-note))
	    (rmoveto score (+ (accidental-breathing-space score)
			      (if (or (<= line -2) (<= 10 line))
				  leger-len
				0))
		     0)))
	(setf nx0 (scr-x score))	;assuming size=1.0 during housing pass!
	(setf nx1 (+ nx0 (rx note)))

	(if (/= nsx 0.0)
	    (progn
	      (if (not (note-sign note)) (rmoveto score nsx 0))
	      (if (sundry-p nh)
		  (display nh note score rest)
		(show score nh :size (* (size score) (note-head-size score)))))
	  (if (sundry-p nh)
	      (display nh note score rest)
	    (show score nh)))
	
	(when (plusp dots)
	  (rmoveto score 
		   (+ (dot-distance score) (if (/= nsx 0.0) (* .3 (note-head-size score)) 0.0))
		   (if (oddp line)
		       0		;in a space so no fixup needed
		     sls))
	  (if (= dots 1)
	      (simple-show score %dot)
	    (show score (apply #'glyphs (loop for dot from 0 below dots
					 collect %dot
					 if (< dot (1- dots)) collect %space)))))
	
	(let* ((stem-dir (or (audible-stem-direction note)
			     (if (>= line 4) 
				 :down?
			       :up?)))
	       (stem-up (member stem-dir '(:up :up?)))
	       (x0 (+ x0-note dx-note
		      (if (and (note-sign note) (zerop (audible-collision note)))
			  (+ (g-rx (note-sign note))
			     (if (or (<= line -2) (<= 10 line))
				 leger-len
			       0)
			     (accidental-breathing-space score))
			0)))
	       (y0 (+ (box-y0 note) dy-note))
	       (need-stem (and (not chord)
			       (not (eq (beamed note) t))
			       (not (eq (audible-stem-direction note) :none))
			       (< (or (head-quarters note) (quarters note)) 4)))
	       (y1 (if need-stem (+ (get-stem-end score note stem-up rest)
				    (or (stem-dy note) 0)))))
	  
	  (when need-stem
	    (if (and (/= nsx 0.0) (not stem-up)) (incf x0 nsx))
	    (draw-stem score stem-up x0 (+ y0 (note-head-y0-offset note stem-up)) y1 (g-rx nh) (stem-mark note))
	    (setf (audible-stem-direction note) stem-dir)
	    (when (plusp nflags) (draw-flags score nflags stem-up y1 x0))
	    (if (member :just-fooling rest)
		(setf (stem-end note) (box-y1 score))
	      (setf (stem-end note) y1)))
	  
	  (if (marks note) (apply #'display-note-slurs-and-brackets note score rest))
	  
	  (if (get-tremolo note) 
	      (let* ((trem (get-tremolo note))
		     (type (tremolo-type trem)))
		(if (not type)
		    (draw-tremolo score note
				  (or (tremolo-slashes trem)
				      (- 3 nflags))
				  (paranoid trem)
				  (if (< (quarters note) 4)
				      (if stem-up
					  (+ x0 (- (note-head-width score) (half-stem-width score)))
					(+ x0 (half-stem-width score)))
				    (+ x0 (* .5 (g-rx %wholenote))))
				  y0 y1 stem-dir
				  (if (and (measured trem)
					   (zerop nflags))
				      .15
				    0))
		  (if (and (eq type :left)
			   (tremolo-note trem))
		      (draw-two-note-tremolo score note trem)))))
	  
	  (when (or (< line -1) (> line 9))
	    (let* ((x1 (+ x0 (rx note)))
		   (yline (if (oddp line) 
			      (if (minusp line) 
				  (+ y0 sls)
				(- y0 sls))
			    y0))
		   (factor (if (minusp line) 2 -2))
		   (lines (if (oddp line) 
			      (if (plusp line) 
				  (floor (- line 9) 2)
				(floor (- (abs line) 1) 2))
			    (if (plusp line)
				(floor (- line 8) 2)
			      (floor (abs line) 2))))
	           (special-leger-fixup-case (and need-stem (/= nsx 0.0) (member stem-dir '(:down :down?))))
		   (lx0 (if special-leger-fixup-case x0 (+ x0 nsx)))
		   (lx1 (if special-leger-fixup-case (- x1 nsx) x1)))
	      (setf (line-width score) (leger-line-width score))

	      #-KCL (loop for i from 0 below lines and y1 from yline by (* factor sls) do
		      (just-line score (- lx0 leger-len) y1 (+ lx1 leger-len)))
	      #+KCL (do ((i 0 (1+ i))
			 (y1 yline (+ y1 (* factor sls))))
			((>= i lines))
		      (just-line score (- lx0 leger-len) y1 (+ lx1 leger-len)))
	      
	      (simple-draw score)
	      (setf (line-width score) 0))))))

    (* .5 (+ nx0 nx1))))

#|
;;; note-head size tests
(cmn (note-head-size .75) staff treble f5 w fs5 h c5 w cs5 h f4 w fs4 h c4 w cs4 h)
(cmn (note-head-size 1.5) staff treble f5 w fs5 h c5 w cs5 h f4 w fs4 h c4 w cs4 h)
(cmn (note-head-size 1.5) staff treble f5 e fs5 e c5 e cs5 e f4 e fs4 e c4 e cs4 e)
(cmn (note-head-size 1.5) staff treble c6 e c6 e c6 q a3 q a3 e a3 e)
(cmn (note-head-size 1.5) staff treble (chord (notes c5 g5 c6) h) (chord (notes a3 d4) h) 
  (chord (notes c5 c6) e) (chord (notes c5 c6) e) (chord (notes a3 e4) e) (chord (notes a3 e4) e))
(cmn (note-head-size 1.5) (grace-note-size 1.0) staff treble (g4 h (appogiatura (notes bn4 cs5))) 
  (bn4 e stem-up) (cs5 e stem-up) g4 h)
|#

(defparameter note-walls '(.05 0 .05 0))
(defparameter note-fences '(.05 0 .05 0))
(defparameter note-expanders '(2 0 3 0))

(defmethod house ((note note) score)
  (let ((old-output (output score))
	(note-center 0))
    (clear-server score)
    (setf note-center (display note nil score :just-fooling))
    (setf (output score) old-output)
    (when (not (eq (visible-justification note) :none))
      (copy-bounding-box score note)
      (setf (center note) note-center)
      (setf (walls note) note-walls)
      (setf (fences note) note-fences)
      (if (not (expanders note)) 
	  (setf (expanders note) note-expanders)))))


(defmacro define-note (name c-class p-class &optional octave sign)
  `(progn
     (defparameter ,name (make-instance 'write-protected-note :cclass ,c-class :pitch ,p-class :octave ,octave :sign ,sign))
     (defun ,name (&rest objects)
       (let ((new-note (make-note 
			:cclass ,c-class
			:pitch ,p-class
			:octave ,octave
			:sign ,sign)))
	 (apply #'ur-note new-note objects)))))


(define-note CF0 0 -1 0 flat) (define-note C0 0 0 0) (define-note CN0 0 0 0 natural) (define-note CS0 0 1 0 sharp)
(define-note DF0 1 1 0 flat)  (define-note D0 1 2 0) (define-note DN0 1 2 0 natural) (define-note DS0 1 3 0 sharp)
(define-note EF0 2 3 0 flat)  (define-note E0 2 4 0) (define-note EN0 2 4 0 natural) (define-note ES0 2 5 0 sharp)
(define-note FF0 3 4 0 flat)  (define-note F0 3 5 0) (define-note FN0 3 5 0 natural) (define-note FS0 3 6 0 sharp)
(define-note GF0 4 6 0 flat)  (define-note G0 4 7 0) (define-note GN0 4 7 0 natural) (define-note GS0 4 8 0 sharp)
(define-note AF0 5 8 0 flat)  (define-note A0 5 9 0) (define-note AN0 5 9 0 natural) (define-note AS0 5 10 0 sharp)
(define-note BF0 6 10 0 flat) (define-note B0 6 11 0)(define-note BN0 6 11 0 natural)(define-note BS0 6 12 0 sharp)

(define-note CF1 0 -1 1 flat) (define-note C1 0 0 1) (define-note CN1 0 0 1 natural) (define-note CS1 0 1 1 sharp)
(define-note DF1 1 1 1 flat)  (define-note D1 1 2 1) (define-note DN1 1 2 1 natural) (define-note DS1 1 3 1 sharp)
(define-note EF1 2 3 1 flat)  (define-note E1 2 4 1) (define-note EN1 2 4 1 natural) (define-note ES1 2 5 1 sharp)
(define-note FF1 3 4 1 flat)  (define-note F1 3 5 1) (define-note FN1 3 5 1 natural) (define-note FS1 3 6 1 sharp)
(define-note GF1 4 6 1 flat)  (define-note G1 4 7 1) (define-note GN1 4 7 1 natural) (define-note GS1 4 8 1 sharp)
(define-note AF1 5 8 1 flat)  (define-note A1 5 9 1) (define-note AN1 5 9 1 natural) (define-note AS1 5 10 1 sharp)
(define-note BF1 6 10 1 flat) (define-note B1 6 11 1)(define-note BN1 6 11 1 natural)(define-note BS1 6 12 1 sharp)

(define-note CF2 0 -1 2 flat) (define-note C2 0 0 2) (define-note CN2 0 0 2 natural) (define-note CS2 0 1 2 sharp)
(define-note DF2 1 1 2 flat)  (define-note D2 1 2 2) (define-note DN2 1 2 2 natural) (define-note DS2 1 3 2 sharp)
(define-note EF2 2 3 2 flat)  (define-note E2 2 4 2) (define-note EN2 2 4 2 natural) (define-note ES2 2 5 2 sharp)
(define-note FF2 3 4 2 flat)  (define-note F2 3 5 2) (define-note FN2 3 5 2 natural) (define-note FS2 3 6 2 sharp)
(define-note GF2 4 6 2 flat)  (define-note G2 4 7 2) (define-note GN2 4 7 2 natural) (define-note GS2 4 8 2 sharp)
(define-note AF2 5 8 2 flat)  (define-note A2 5 9 2) (define-note AN2 5 9 2 natural) (define-note AS2 5 10 2 sharp)
(define-note BF2 6 10 2 flat) (define-note B2 6 11 2)(define-note BN2 6 11 2 natural)(define-note BS2 6 12 2 sharp)

(define-note CF3 0 -1 3 flat) (define-note C3 0 0 3) (define-note CN3 0 0 3 natural) (define-note CS3 0 1 3 sharp)
(define-note DF3 1 1 3 flat)  (define-note D3 1 2 3) (define-note DN3 1 2 3 natural) (define-note DS3 1 3 3 sharp)
(define-note EF3 2 3 3 flat)  (define-note E3 2 4 3) (define-note EN3 2 4 3 natural) (define-note ES3 2 5 3 sharp)
(define-note FF3 3 4 3 flat)  (define-note F3 3 5 3) (define-note FN3 3 5 3 natural) (define-note FS3 3 6 3 sharp)
(define-note GF3 4 6 3 flat)  (define-note G3 4 7 3) (define-note GN3 4 7 3 natural) (define-note GS3 4 8 3 sharp)
(define-note AF3 5 8 3 flat)  (define-note A3 5 9 3) (define-note AN3 5 9 3 natural) (define-note AS3 5 10 3 sharp)
(define-note BF3 6 10 3 flat) (define-note B3 6 11 3)(define-note BN3 6 11 3 natural)(define-note BS3 6 12 3 sharp)

(define-note CF4 0 -1 4 flat) (define-note C4 0 0 4) (define-note CN4 0 0 4 natural) (define-note CS4 0 1 4 sharp)
(define-note DF4 1 1 4 flat)  (define-note D4 1 2 4) (define-note DN4 1 2 4 natural) (define-note DS4 1 3 4 sharp)
(define-note EF4 2 3 4 flat)  (define-note E4 2 4 4) (define-note EN4 2 4 4 natural) (define-note ES4 2 5 4 sharp)
(define-note FF4 3 4 4 flat)  (define-note F4 3 5 4) (define-note FN4 3 5 4 natural) (define-note FS4 3 6 4 sharp)
(define-note GF4 4 6 4 flat)  (define-note G4 4 7 4) (define-note GN4 4 7 4 natural) (define-note GS4 4 8 4 sharp)
(define-note AF4 5 8 4 flat)  (define-note A4 5 9 4) (define-note AN4 5 9 4 natural) (define-note AS4 5 10 4 sharp)
(define-note BF4 6 10 4 flat) (define-note B4 6 11 4)(define-note BN4 6 11 4 natural)(define-note BS4 6 12 4 sharp)

(define-note CF5 0 -1 5 flat) (define-note C5 0 0 5) (define-note CN5 0 0 5 natural) (define-note CS5 0 1 5 sharp)
(define-note DF5 1 1 5 flat)  (define-note D5 1 2 5) (define-note DN5 1 2 5 natural) (define-note DS5 1 3 5 sharp)
(define-note EF5 2 3 5 flat)  (define-note E5 2 4 5) (define-note EN5 2 4 5 natural) (define-note ES5 2 5 5 sharp)
(define-note FF5 3 4 5 flat)  (define-note F5 3 5 5) (define-note FN5 3 5 5 natural) (define-note FS5 3 6 5 sharp)
(define-note GF5 4 6 5 flat)  (define-note G5 4 7 5) (define-note GN5 4 7 5 natural) (define-note GS5 4 8 5 sharp)
(define-note AF5 5 8 5 flat)  (define-note A5 5 9 5) (define-note AN5 5 9 5 natural) (define-note AS5 5 1 5 sharp)
(define-note BF5 6 10 5 flat) (define-note B5 6 11 5)(define-note BN5 6 11 5 natural)(define-note BS5 6 12 5 sharp)

(define-note CF6 0 -1 6 flat) (define-note C6 0 0 6) (define-note CN6 0 0 6 natural) (define-note CS6 0 1 6 sharp)
(define-note DF6 1 1 6 flat)  (define-note D6 1 2 6) (define-note DN6 1 2 6 natural) (define-note DS6 1 3 6 sharp)
(define-note EF6 2 3 6 flat)  (define-note E6 2 4 6) (define-note EN6 2 4 6 natural) (define-note ES6 2 5 6 sharp)
(define-note FF6 3 4 6 flat)  (define-note F6 3 5 6) (define-note FN6 3 5 6 natural) (define-note FS6 3 6 6 sharp)
(define-note GF6 4 6 6 flat)  (define-note G6 4 7 6) (define-note GN6 4 7 6 natural) (define-note GS6 4 8 6 sharp)
(define-note AF6 5 8 6 flat)  (define-note A6 5 9 6) (define-note AN6 5 9 6 natural) (define-note AS6 5 10 6 sharp)
(define-note BF6 6 10 6 flat) (define-note B6 6 11 6)(define-note BN6 6 11 6 natural)(define-note BS6 6 12 6 sharp)

(define-note CF7 0 -1 7 flat) (define-note C7 0 0 7) (define-note CN7 0 0 7 natural) (define-note CS7 0 1 7 sharp)
(define-note DF7 1 1 7 flat)  (define-note D7 1 2 7) (define-note DN7 1 2 7 natural) (define-note DS7 1 3 7 sharp)
(define-note EF7 2 3 7 flat)  (define-note E7 2 4 7) (define-note EN7 2 4 7 natural) (define-note ES7 2 5 7 sharp)
(define-note FF7 3 4 7 flat)  (define-note F7 3 5 7) (define-note FN7 3 5 7 natural) (define-note FS7 3 6 7 sharp)
(define-note GF7 4 6 7 flat)  (define-note G7 4 7 7) (define-note GN7 4 7 7 natural) (define-note GS7 4 8 7 sharp)
(define-note AF7 5 8 7 flat)  (define-note A7 5 9 7) (define-note AN7 5 9 7 natural) (define-note AS7 5 10 7 sharp)
(define-note BF7 6 10 7 flat) (define-note B7 6 11 7)(define-note BN7 6 11 7 natural)(define-note BS7 6 12 7 sharp)

(define-note CF8 0 -1 8 flat) (define-note C8 0 0 8) (define-note CN8 0 0 8 natural) (define-note CS8 0 1 8 sharp)
(define-note DF8 1 1 8 flat)  (define-note D8 1 2 8) (define-note DN8 1 2 8 natural) (define-note DS8 1 3 8 sharp)
(define-note EF8 2 3 8 flat)  (define-note E8 2 4 8) (define-note EN8 2 4 8 natural) (define-note ES8 2 5 8 sharp)
(define-note FF8 3 4 8 flat)  (define-note F8 3 5 8) (define-note FN8 3 5 8 natural) (define-note FS8 3 6 8 sharp)
(define-note GF8 4 6 8 flat)  (define-note G8 4 7 8) (define-note GN8 4 7 8 natural) (define-note GS8 4 8 8 sharp)
(define-note AF8 5 8 8 flat)  (define-note A8 5 9 8) (define-note AN8 5 9 8 natural) (define-note AS8 5 10 8 sharp)
(define-note BF8 6 10 8 flat) (define-note B8 6 11 8)(define-note BN8 6 11 8 natural)(define-note BS8 6 12 8 sharp)

(defun place-of-note-on-clef (note clef) ;note is '(pitch-class octave sign bare-note-name ...)
  (let* ((clef-relative-octave (- (octave note)
				  (floor (clef-base-pitch clef) 12)))
	 (cclass (cclass note)))
    ;; clef goes from -1 to 9 -- above or below we need leger lines or octave signs
    (+ (* 7 clef-relative-octave) 
       (- cclass (clef-base-line-note clef)))))

(defun place-of-note-given-note (known-note unknown-note)
  ;; given line number of known-note, find (and set) line number of unknown-note
  (let ((line (note-line known-note))
	(oct-off (- (octave unknown-note) (octave known-note)))
	(cls-off (- (cclass unknown-note) (cclass known-note))))
    (+ line cls-off (* 7 oct-off))))




;;;
;;; ----------------    chords
;;;

(defclass chord (audible rhythm staff-relative-mixin score-object)
  ((data :initarg :data :initform nil :accessor chord-data)
   (note-head :initarg :note-head :initform nil :accessor note-head)
   (stem-dy :initarg :stem-dy :initform nil :accessor stem-dy)
   (stem-mark :initarg :stem-mark :initform nil :accessor stem-mark)))

(defmethod chord-p ((obj t)) nil)
(defmethod chord-p ((obj chord)) t)

(defmethod descry ((chord chord) &optional stream controller)
  (format stream "(chord~A~A~A~A)"
	  (if (next-method-p) (call-next-method chord stream (or controller chord)) "")
	  (if (note-head chord) (format nil " :note-head :~(~A~)" (note-head chord)) "")
	  (if (stem-dy chord) (format nil " :stem-dy ~A" (stem-dy chord)) "")
	  (if (chord-data chord) (format nil " :data (list~{~%              ~A~})" 
					 (loop for note in (chord-data chord) collect (descry note))) 
	    "")))

(defmethod brief-identify ((chord chord))
  (format nil "(chord ~A~A~A)"
	  (if (not (halve-duration chord))
	      (rhythm-name (quarters chord))
	    (rhythm-name (* 2 (quarters chord))))
	  (if (onset chord) (format nil " (onset ~A)" (fratify (onset chord))) "")
	  (if (chord-data chord)
	      (format nil " (notes~{ ~A~})"
		      (loop for note in (chord-data chord)
		       collect (note-name note))))
	    ""))

(defmethod identify ((chord chord))
  (format nil "(chord ~A~A~A~A~A~A~A~A~A~A~A~A)"
	  (if (not (halve-duration chord))
	      (rhythm-name (quarters chord))
	    (rhythm-name (* 2 (quarters chord))))
	  (if (onset chord) (format nil " (onset ~A)" (fratify (onset chord))) "")
	  (if (chord-data chord)
	      (format nil " (notes~{ ~A~})"
		      (loop for note in (chord-data chord)
		       collect (if (or (marks note) (store-data note) (ties note))
				   (identify note) 
				 (note-name note))))
	    "")
	  (if (and (note-head chord)
		   (not (eq (note-head chord) :normal)))
	      (format nil " (note-head :~(~A~))" (note-head chord))
	    "")
	  (if (eq (stem-direction chord) :up) " stem-up"
	    (if (eq (stem-direction chord) :down) " stem-down"
	      (if (eq (stem-direction chord) :none) " no-stem"
		"")))
	  (if (stem-dy chord) (format nil " (stem-dy ~A)" (stem-dy chord)) "")
	  (if (audible-tie-direction chord) (format nil " (tie-direction :~(~A~))" (audible-tie-direction chord)) "")
	  (if (audible-slur-direction chord) (format nil " (slur-direction :~(~A~))" (audible-slur-direction chord)) "")
	  (if (store-data chord) (identify-store-data chord) "")
	  (if (ties chord) (format nil "~{ ~(~A~)~}" (sorted-ties (ties chord))) "")
	  (if (get-tremolo chord) (identify (get-tremolo chord)) "")
	  (the-usual-suspects chord)))

(defmethod copy ((chord chord) &optional object)
  (let ((new-chord (if (not object) (make-chord)
		     (if (write-protected object) (copy object)
		       object))))
    (if (chord-data chord) (setf (chord-data new-chord) (loop for note in (chord-data chord) collect (copy note))))
    (setf (note-head new-chord) (note-head chord))
    (if (next-method-p) (call-next-method chord new-chord))
    new-chord))

(defun copy-chord-for-ties (chord)
  (let ((new-chord (make-chord)))
    (setf (dots new-chord) (dots chord))
    (setf (flags new-chord) (flags chord))
    (setf (quarters new-chord) (quarters chord))
    (setf (beamed new-chord) (beamed chord))
    (setf (octaved new-chord) (octaved chord))
    (setf (note-head new-chord) (note-head chord))
    (setf (stem-dy new-chord) (stem-dy chord))
    (setf (stem-mark new-chord) (stem-mark chord))
    (if (get-tremolo chord) (setf (get-tremolo new-chord) (copy (get-tremolo chord))))
    (setf (audible-stem-direction new-chord) (audible-stem-direction chord))
    (setf (audible-tie-direction new-chord) (audible-tie-direction chord))
    (setf (audible-slur-direction new-chord) (audible-slur-direction chord))
    (setf (chord-data new-chord) (loop for note in (chord-data chord) collect (copy note)))
    (loop for note in (chord-data new-chord) do (setf (note-sign note) nil))
    new-chord))

(defmethod short-note ((chord chord) &key x0)
  (make-chord :x0 (or x0 (box-x0 chord)) :x1 (box-x1 chord) :y0 (box-y0 chord) :y1 (box-y1 chord)
	      :staff-y0 (%staff-y0 chord) :stem-direction (audible-stem-direction chord)
	      :tie-direction (audible-tie-direction chord) :slur-direction (audible-slur-direction chord)
	      :center (center chord) :ties (if (ties chord) t nil) :quarters (quarters chord)
	      :stem-end (stem-end chord) 
	      :data (loop for note in (chord-data chord) collect (short-note note))))

(defmethod add-to-marks ((chord chord) objects) (setf (marks chord) (append (marks chord) objects)))
(defmethod add-to-ties ((chord chord) object) (push object (ties chord)))
(defmethod add-to-slurs ((chord chord) objects) (push objects (slurs chord)))

(defun chord (&rest objects)
  (let ((notes nil)
	(new-chord (make-chord))
	(object nil))
    (setf *cmn-owning-object* new-chord)
    (loop while objects do
      (setf object (pop objects))
      (setf *cmn-object* object)
      (when object
	(if (self-acting-p object)
	    (funcall (action object) new-chord (arguments object))
	  (if (note-p object)
	      (push (if (write-protected object) (copy object) object) notes)
	    (if (rhythm-p object)
		(rcopy object new-chord)
	      (if (score-object-p object)
		  (if (or (sundry-p object) (pause-p object) (dynamics-p object) (glyph-list-p object))
		      (add-to-marks new-chord (list (if (write-protected object) (copy object) object)))
		    (copy object new-chord))
		(if (score-object-list-p object)
		    (setf objects (append (disgorge object) objects))
		  (if (text-p object)
		      (add-to-marks new-chord (list object))
		    (cmn-warn "odd argument to chord: ~A" object)))))))))
    (setf (chord-data new-chord) (nreverse notes))
    (if (note-head new-chord)
	(loop for note in (chord-data new-chord) do (setf (note-head note) (note-head new-chord))))
    (when (and (not (quarters new-chord))
	       (not (duration new-chord)))
      (let ((fnote (or (find-if-not #'null (chord-data new-chord) :key #'quarters)
		       (first (chord-data new-chord)))))
	(setf (quarters new-chord) (quarters fnote))
	(setf (odb-duration new-chord) (odb-duration fnote))
	(setf (odb-onset new-chord) (or (odb-onset new-chord) (odb-onset fnote)))
	(setf (dots new-chord) (dots fnote))
	(setf (flags new-chord) (flags fnote))))
    (loop for note in (chord-data new-chord) do
      (when (not (odb-onset note))	;might have been a notes list (was quarters -- defaults to 1)
	(setf (quarters note) (quarters new-chord))
	(setf (odb-duration note) (odb-duration new-chord))
	(setf (odb-onset note) (odb-onset new-chord)))
      (setf (dots note) (dots new-chord))
      (setf (flags note) (flags new-chord)))
    (setf (octaved new-chord) (or (octaved new-chord) octave-stack))
    (cmn-tick-pipe new-chord)
    new-chord))

(defun notes (&rest ns) (engorge ns))

(defmethod notify ((chord chord) &optional objects)
  (apply #'chord chord objects))

(defmethod (setf sign) (val (chord chord))
  (loop for note in (chord-data chord) do
    (setf (note-sign note) val)))

(defmethod sign ((chord chord))
  (let ((note (find-if-not #'null (chord-data chord) :key #'sign)))
    (and note (note-sign note))))

(defmethod setf-rhythm ((chord chord) ur-val)
  (let ((val (fratify ur-val)))
    (setf (odb-duration chord) val)
    (setf (dots chord) (how-many-dots val))
    (setf (flags chord) (how-many-flags val))
    (setf (quarters chord) val)
    (loop for note in (chord-data chord) do
      (setf-rhythm note val)
      (setf (dots note) 0))))

(defun displace-chord (ordered-chord-1 lines-1 stem-direction)
  (let* ((stem-up (member stem-direction '(:up :up?)))
	 (normal (if stem-up :left :right))
	 (other (if stem-up :right :left))
	 (displaced normal)
	 (last-line nil)
	 (ordered-chord (if stem-up (reverse ordered-chord-1) ordered-chord-1))
	 (lines (if stem-up (reverse lines-1) lines-1))
	 (chords (loop for note in ordered-chord and 
		  line in lines
		  do (progn
		       (if (and last-line 
				(eq displaced normal)
				(or (= line last-line) 
				    (and stem-up
					 (= line (1+ last-line)))
				    (and (not stem-up)
					 (= (1+ line) last-line))))
			   (setf displaced other)
			 (setf displaced normal))
		       (setf last-line line))
		   ;; eventually we should probably respell the thing if on the same line and not the same sign
		  collect (list 0 displaced line (note-sign note) note))))
    (if stem-up (nreverse chords) chords)))

(defun accident-chord (chord stem-direction)
  ;; chord is a list of notes made up of "0", note-head displacements, line numbers, and the signs [original note values]
  ;; here we decide where to put the accidentals so that they don't collide with each other or with the note heads.
  ;; we return the old list with a new first element = accidental displacement levels (i.e. sign motion backwards)
  ;; problem is that the accidental signs are different sizes.
  ;; basic algorithm is that a seventh gives a room on any given layer of accidentals, layers can be any depth
  ;; a "seventh" is (oddly enough) a difference of 6 in line numbers
  ;; return the max (left) displacement and the nominal true-center displacement (where the note heads go modulo seconds)
  ;; each of the possible cases of collision has a different level map shape (i.e. flats go down very little, but
  ;;   sharps go down a lot, so we have different ways of deciding when a level is free)
  (let ((got-signs (some #'fourth chord)))
    (if (not got-signs)
	chord
      (let* ((all-done nil)
	     (level 1)
	     (notes (length chord))
	     (level-map (make-array notes :element-type 'short-float :initial-element 1.0))
	     (lines (make-array notes :element-type 'fixnum :initial-contents (map 'list #'third chord))))
	
	;; map out collisions with note heads on the "wrong" side of a downward stem (getting in accidental's way)
	(loop for note-data in chord and i from 0 do
	  (if (and (member stem-direction '(:down :down?))
		   (eq (second note-data) :left))
	      (loop for k from 0 below notes do
	        (let ((diff (abs (- (aref lines k) (aref lines i)))))
		  (if (or (= diff 3) (= diff -4))
		      (setf (aref level-map k) (max (aref level-map k) 1.5))
		    (if (< diff 3)
			(setf (aref level-map k) 2.0)))))))
	
	;; an accidental fits into a level if the current level-map is less than or equal to the current level
	(loop until all-done do		;keep filling levels until all signs dealt with 
	  (let ((unset-signs nil))
	    (loop for note-data in chord and i from 0 do
	      (let ((sign (fourth note-data)))
		(when (and sign		;there's an accidental
			   (zerop (first note-data))) ; that hasn't yet been placed
		  (if (<= (aref level-map i) level) ; and we have room on this level
		      (progn
			(setf (first note-data) level)
			(if (eq sign flat)
			    (loop for k from 0 below notes do
			      (let ((diff (- (aref lines k) (aref lines i))))
				(if (or (= diff -3) (= diff -4) (= diff -5))
				    (setf (aref level-map k) (max (aref level-map k) (+ level .5)))
				  (if (< (abs diff) 5)
				      (setf (aref level-map k) (max (aref level-map k) (+ level 1.0)))))))
			  (if (eq sign sharp)
			      (loop for k from 0 below notes do
				(let ((diff (abs (- (aref lines k) (aref lines i)))))
				  (if (< diff 6)
				      (setf (aref level-map k) (max (aref level-map k) (+ level 1.0))))))
			    (if (eq sign natural)
				(loop for k from 0 below notes do
				  (let ((diff (- (aref lines k) (aref lines i))))
				    (if (or (= diff -4) (= diff -5))
					(setf (aref level-map k) (max (aref level-map k) (+ level .5)))
				      (if (< (abs diff) 6)
					  (setf (aref level-map k) (max (aref level-map k) (+ level 1.0)))))))
			      (if (eq sign double-sharp)
				  (loop for k from 0 below notes do
				    (let ((diff (abs (- (aref lines k) (aref lines i)))))
				      (if (< diff 4)
					  (setf (aref level-map k) (max (aref level-map k) (+ level 1.0))))))
				(if (eq sign double-flat)
				    (loop for k from 0 below notes do
				      (let ((diff (- (aref lines k) (aref lines i))))
					(if (or (= diff 3) (= diff 4) (= diff -3))
					    (setf (aref level-map k) (max (aref level-map k) (+ level 1.5)))
					  (if (< (abs diff) 5)
					      (setf (aref level-map k) (max (aref level-map k) (+ level 2.0)))))))))))))
		    (setf unset-signs t)))))
	    (setf all-done (not unset-signs))
	    (incf level .5)))
	chord))))

(defmethod display ((chord  chord) container score &rest rest)
  (declare (ignore container))
  ;; chord is a list of notes -- each note is a list just like place-note takes.
  ;; return the x0 loc relative to px0 of the main notes (for multi-staff alignment) and the bounding box of the chord
  (let* ((stf-y0 (%staff-y0 chord))
	 (lines (line chord))
	 (dx-chord (if (member :just-fooling rest) 0 (dxy-dx chord)))
	 (dy-chord (if (member :just-fooling rest) 0 (dxy-dy chord)))
	 (x0 (+ (box-x0 chord) dx-chord))
	 (y0 (+ stf-y0 dy-chord))
	 ;; stem direction determined by note furthest from staff (in LINES)
	 (max-line (loop for line in lines maximize line))
	 (min-line (loop for line in lines minimize line))
	 ;; "4" is the center line of the staff
	 (stem-direction (or (audible-stem-direction chord)
			     (if (> (abs (- max-line 4)) (- 4 min-line)) :down? :up?)))
	 (stem-up (member stem-direction '(:up :up?)))
	 (stem-down (not stem-up))
	 (stem-start (if stem-down max-line min-line))
	 (stem-last-note (if stem-down min-line max-line))
	 (stem-extra (if (member :grace-note rest) 4 6)) ; i.e. 3 staff spaces = 6 "lines" extra on stem after last note
	 (stem-end-1 (if stem-down 
			 (- stem-last-note stem-extra) 
		       (+ stem-last-note stem-extra)))
	 ;; so stem goes from line STEM-START to STEM-END-1 relative to current staff
	 ;; this has not yet taken flags or leger lines into account (may be lengthened a lot)
	 ;; if max-line is > 10 or min-line < -2 we have leger lines, and the stem has to reach the mid line (4)
	 (stem-end-2 (if (and stem-up 
			      (< min-line -2))
			 (max stem-end-1 4)
		       (if (and stem-down
				(> max-line 10))
			   (min stem-end-1 4)
			 stem-end-1)))
	 ;; and room for flags...
	 
	 (true-head (get-note-head (quarters chord) (note-head (first (chord-data chord)))))
	 (nflags (or (flags chord) 0))
	 (ntrem (or (and (get-tremolo chord) (or (tremolo-slashes (get-tremolo chord)) (- 3 nflags))) 0))
	 (stem-end (if (and (< nflags 2) (zerop ntrem))
		       stem-end-2
		     (if stem-up
			 (max stem-end-2 (+ stem-end-1 nflags ntrem))
		       (min stem-end-2 
			    (- stem-end-1 nflags ntrem 
			       (if (and (plusp nflags)
					(> (note-head-size score) 1.1))
				   (* .5 (- (note-head-size score) 1.1))
				 0))))))
		       ;; (min stem-end-2 (- stem-end-2 nflags ntrem)) ; is this right?  It looks silly...
	 
	 (ordered-chord (sort (copy-list (chord-data chord))
			      #'(lambda (note1 note2)
				  (> (pitch-of note1) (pitch-of note2)))))
	 
	 (ordered-lines (loop for note in ordered-chord collect (note-line note)))
	 ;; looks like best way to plot a chord is to start at the top and work down.
	 ;; we need to make space for accidentals and noteheads on the left (seconds)
	 ;; there can be duplicate notes, or notes on the same staff line (or space), 
	 ;; and dots may have to be displaced down rather than up.
	 (displaced-chord (displace-chord ordered-chord ordered-lines stem-direction))
	 (accidentals (accident-chord displaced-chord stem-direction))
	 (normal-note-side (if stem-up :left :right))
	 (other-note-side (if (eq normal-note-side :right) :left :right))
	 (head-width (g-rx true-head))
	 (nsx (* .3 (- 1.0 (note-head-size score))))
	 (stem-half-width (if (>= (quarters chord) 4) 0.0 (half-stem-width score)))
	 (max-displacement (loop for note in accidentals maximize (first note)))
         (sign-displacement-right (* .25 max-displacement))
	 (sign-breathing-room (if (zerop max-displacement) 0.0 (note-breathing-space score)))
	 (mark-displacement-right (if (marks chord)
				      (if (find-if #'arrow-p (marks chord))
					  .2
					(if (find-if #'arpeggio-p (marks chord))
					    .25
					  0))
				    0))
	 (normal-head-x0 (+ x0 
			    sign-displacement-right sign-breathing-room 
			    mark-displacement-right
			    (if stem-up
				0.0
			      (+ stem-half-width
				 (if (and (zerop sign-displacement-right)
					  (find other-note-side accidentals :key #'second))
				     head-width
				   0.0)))))
	 (other-head-x0 (+ normal-head-x0
			   (if (>= (quarters chord) 4)
			       (if stem-up
				   (- head-width nsx .075)
				 (- (+ .075 nsx) head-width))
			     (if stem-up
				 (- head-width nsx stem-half-width) ;used to be (+ stem-half-width head-width)
			       (- nsx head-width)))))
	 (other-note-x0-if-in-use-and-on-right (and (eq other-note-side :right) 
						    (find :right displaced-chord :key #'second)
						    other-head-x0)))

    (setf (audible-stem-direction chord) stem-direction)
    
    (setf (box-y0 chord) (+ (%staff-y0 chord) (* (staff-line-separation score) min-line)))
    
    (when (and (not (member :just-fooling rest))
	       (not (member :just-do-it rest)))
      (if (beams chord) (display (beams chord) chord score))
      (if (ties chord) 
	  (loop for tie in (ties chord) do 
	    (display-tied-chord score chord tie
				accidentals 
				(min normal-head-x0 other-head-x0)
				(max normal-head-x0 other-head-x0)
				stem-direction ordered-lines min-line max-line))))
    (if (marks chord) (apply #'display-marks chord score rest))
    
    ;; put in stem and flags (if not beaming)
    
    (if (eq (beamed chord) t)
	(if (marks chord) (apply #'display-note-slurs-and-brackets chord score rest))
      (let* ((bm-x0 (if stem-up 
			(+ other-head-x0 nsx stem-half-width)
		      (+ normal-head-x0 nsx stem-half-width)))
	     (fl-x0 (if stem-up 
			(- bm-x0 (- (note-head-width score) (half-stem-width score)) )
		      (- bm-x0 (half-stem-width score))))
	     ;; .275 is nominal x offset, .0125 is half stem width => .2875
	     (sy0 (+ y0 
		     (* stem-start (staff-line-separation score)) 
		     (if stem-up 
			 (half-stem-width score) 
		       (- (half-stem-width score)))))
	     (sy1 (+ y0 (* stem-end (staff-line-separation score)))))
	(when (and (< (quarters chord) 4) 
		   (not (eq (audible-stem-direction chord) :none)))
	  (optimize-stem score bm-x0 sy0 (+ sy1 (or (stem-dy chord) 0)) (stem-width score))
	  (if (stem-mark chord) (funcall (stem-mark chord) score bm-x0 sy0 sy1))
	  (setf (stem-end chord) sy1)
	  (when (plusp nflags) (draw-flags score nflags (member stem-direction '(:up :up?)) sy1 fl-x0)))
	
	(if (marks chord) (apply #'display-note-slurs-and-brackets chord score rest))
	
	(when (get-tremolo chord) 
	  (draw-tremolo score chord
			(or (tremolo-slashes (get-tremolo chord)) 
			    (- 3 nflags))
			(paranoid (get-tremolo chord))
			(if (< (quarters chord) 4)
			    bm-x0
			  (if stem-up
			      (- bm-x0 (* .5 (g-rx %wholenote)))
			    (+ bm-x0 (* .5 (g-rx %wholenote)))))
			(+ (%staff-y0 chord) (* stem-last-note (staff-line-separation score)))
			sy1 stem-direction
			(if (and (measured (get-tremolo chord))
				 (zerop nflags))
			    .15
			  .10)))))
    
    (loop for ur-note in accidentals do	; cons up a fake note to get the note head and (possible) leger lines taken care of
      (let* ((tnote (fifth ur-note))
	     (old-sign (note-sign tnote)))
	(when (not (member :just-fooling rest))
	  (setf (flags tnote) 0)
	  (setf (dots tnote) 0)
	  (setf (quarters tnote) (quarters chord))
	  (setf (note-sign tnote) nil)
	  (setf (box-x0 tnote) (if (eq (second ur-note) normal-note-side) 
				   normal-head-x0 
				 other-head-x0)))
	(apply #'display tnote chord score rest)
	
	(if (not (member :just-fooling rest)) (setf (note-sign tnote) old-sign))))
    
    (loop for note in accidentals do
      (if (fourth note)			;it's got an accidental
	  (progn
	    (moveto score
		    (+ x0 mark-displacement-right nsx
		       (if (and stem-down (eq (second note) :left)) nsx 0)
		       (if (or (and (= (first note) 1.0)
				    (or (> (third note) 9) (< (third note) -1)))
			       (and (= (first note) 2.0)
				    stem-down
				    (eq (second note) :left)
				    (or (> (third note) 9) (< (third note) -1))))
			   (- (leger-line-breathing-space score)) ;try not to collide with the leger line
			 0.0)
		       (* .25 (- max-displacement (first note))))
		    (+ y0 (* (staff-line-separation score) (third note))))
	    (display (fourth note) chord score))))
    
    (let ((last-line 100))
      
      (loop for note in accidentals do
	(let* ((dots (dots chord)))
	  (when (and dots (plusp dots))
	    (let* ((our-line (third note))
		   (collision-above (find (1+ our-line) lines))
		   (collision-below (find (1- our-line) lines)))
	      (moveto score 
		      (+ normal-head-x0 head-width (dot-distance score)
			 (if stem-up
			     (if (eq (second note) :right)
				 (if (and collision-above collision-below
					  (evenp our-line))
				     .35
				   .25)
			       (if (and (eq (second note) :left)
					(or collision-above collision-below))
				   .25
				 0))
			   (if (eq (second note) :left)
			       (if (and collision-above collision-below
					(evenp our-line))
				   .1
				 0)
			     0)))
		      (+ y0 
			 (* (staff-line-separation score) (third note)) 
			 (if (oddp (third note)) 
			     (progn (setf last-line our-line) 0)
			   (if (and collision-above collision-below)
			       (progn (setf last-line our-line) 0)
			     (if (or (and stem-up
					  (= (1- last-line) our-line))
				     (and stem-down
					  ;; this would be easier if we reversed accidentals and went top down
					  (evenp our-line)
					  (or collision-above
					      (and (find (+ our-line 2) lines)
						   (find (+ our-line 3) lines)
						   (not (find (+ our-line 4) lines))))))
				 (progn (setf last-line (1- our-line)) (- (staff-line-separation score)))
			       (progn (setf last-line (1+ our-line)) (staff-line-separation score))))))))
	    (if (= dots 1)
		(simple-show score %dot)
	      (show score (apply #'glyphs (loop for dot from 0 below dots 
					   collect %dot
					   if (< dot (1- dots)) collect %space)))))
	  )))
    
    (values (+ normal-head-x0 (* .5 head-width)) other-note-x0-if-in-use-and-on-right)))

#|
;;; dot placement tests
(cmn (staff treble (chord (notes f4 g4) q.) (chord (notes f4 g4 a4) q.) 
	    (chord (notes e4 f4 g4 a4) q.) (chord (notes e4 f4 g4) q.) 
	    (chord (notes e4 g4 a4) q.) (chord (notes f4 a4 b4) q.) 
	    (chord (notes e4 f4) q.) (chord (notes f4 g4 a4 b4) q.) 
	    (chord (notes e4 f4 a4) q.) (chord (notes f4 g4 b4) q.))
     (staff treble (chord (notes f5 g5) q.) (chord (notes f5 g5 a5) q.) 
	    (chord (notes e5 f5 g5 a5) q.) (chord (notes e5 f5 g5) q.) 
	    (chord (notes e5 g5 a5) q.) (chord (notes f5 a5 b5) q.) 
	    (chord (notes e5 f5) q.) (chord (notes f5 g5 a5 b5) q.) 
	    (chord (notes e5 f5 a5) q.) (chord (notes f5 g5 b5) q.)))

;;; note-head-size tests
(cmn (note-head-size 1.5) staff treble (chord (notes a3 b3) q) (chord (notes a3 b3) e) (chord (notes a3 b3) e) )
(cmn (note-head-size 1.5) staff treble (chord (notes c5 g5 c6) h) (chord (notes c5 d5) e) (chord (notes c5 d5) e) (chord (notes c5 d5) q) )

(cmn staff treble (meter 4 4) (chord (notes a4 e5) (duration 7)))
|#

(defmethod house ((chord chord) score)
  (let ((old-output (output score))
	(chord-center 0)
	(other-note-x0 0))
    (clear-server score)
    (multiple-value-setq
	(chord-center other-note-x0)
      (display chord nil score :just-fooling))
    (setf (output score) old-output)
    (when (not (eq (visible-justification chord) :none))
      (copy-bounding-box score chord)
      (when (ties chord) 
	(if (find-if #'(lambda (tie) 
			 (eq (tie-type tie) :left))
		     (ties chord))
	    (incf (box-x1 chord) .125))
	(if (find-if #'(lambda (tie) 
			 (eq (tie-type tie) :right))
		     (ties chord))
	    (decf (box-x0 chord) .125)))

      (if other-note-x0 (setf (box-x1 chord) (max (box-x1 chord) (+ other-note-x0 .3))))

      (setf (center chord) chord-center)
      (setf (walls chord) note-walls)
      (setf (fences chord) note-fences)
      (if (not (expanders chord))
	  (setf (expanders chord) note-expanders)))))



;;;
;;; ----------------    rests
;;;

(defclass rest-mixin (rhythm-mixin staff-relative-mixin score-object-mixin)
  ((mark :initform nil :initarg :mark :reader rest-mark)
   (line :initform nil :initarg :line :reader rest-line)
   (pause :initform nil :initarg :pause :reader rest-pause)
   (ties :initform nil :initarg :ties :reader ties)
   (tie-direction :initarg :tie-direction :initform nil :reader tie-direction)
   (beamed :initform nil :initarg :beamed :reader beamed)
   (outer-beam :initarg :outer-beam :initform nil :reader outer-beam)
   (store-data :initarg :store-data :initform nil :reader store-data)
   (beams :initarg :beams :initform nil :reader beams)))

(defclass wrest (rest-mixin rhythm score-object)
  ((mark :accessor rest-mark)
   (line :accessor rest-line)
   (pause :accessor rest-pause)
   (ties :accessor ties)
   (tie-direction :accessor tie-direction)
   (beamed :accessor beamed)
   (outer-beam :accessor outer-beam)
   (store-data :accessor store-data)
   (beams :accessor beams)))

(defclass write-protected-rest (write-protect rest-mixin) ())

(defmacro define-rest (name mark dots flags quarters)
  `(progn
     (defvar ,name (make-instance 'write-protected-rest :mark ,mark :dots ,dots :flags ,flags :quarters ,quarters :duration ,quarters))
     (defun ,name (&rest objects) 
       (apply #'ur-rest (make-rest-1 :mark ,mark :dots ,dots :flags ,flags :quarters ,quarters :duration ,quarters) objects))))

(define-rest double-whole-rest       %dblwholerest  0 0 8)
(define-rest dotted-whole-rest       %wholerest     1 0 6)
(define-rest whole-rest              %wholerest     0 0 nil)
(define-rest measure-rest            %measurerest   0 0 nil)
(define-rest repeat-measure          %repsign       0 0 nil)
(define-rest half-rest               %halfrest      0 0 2)
(define-rest dotted-half-rest        %halfrest      1 0 3)
(define-rest quarter-rest            %quarterrest   0 0 1)
(define-rest dotted-quarter-rest     %quarterrest   1 0 3/2)
(define-rest eighth-rest             %eighthrest    0 1 1/2)
(define-rest dotted-eighth-rest      %eighthrest    1 1 3/4)
(define-rest sixteenth-rest          %sixteenthrest 0 2 1/4)
(define-rest dotted-sixteenth-rest   %sixteenthrest 1 2 3/8)
(define-rest thirty-second-rest      %thirtyrest    0 3 1/8)
(define-rest sixty-fourth-rest       %sixtyfourrest 0 4 1/16)
(define-rest one-twenty-eighth-rest  %one28rest     0 5 1/32)

(defun rest-name (rst)
  (let ((rhy (quarters rst)))
    (if (not rhy) 
	(if (eq (rest-mark rst) %repsign) 
	    "repeat-measure" 
	  "whole-rest")			;see above -- whole-rest and measure-rest have nil as quarters field
      (if (= rhy 8) "double-whole-rest"
	(if (= rhy 6) (if (= (dots rst) 1) "dotted-whole-rest" "whole-rest")
	  (if (= rhy 4) "whole-rest"
	    (if (= rhy 2) "half-rest"
	      (if (= rhy 3) "dotted-half-rest"
		(if (= rhy 1) "quarter-rest"
		  (if (= rhy 3/2) "dotted-quarter-rest"
		    (if (= rhy 1/2) "eighth-rest"
		      (if (= rhy 3/4) "dotted-eighth-rest"
			(if (= rhy 1/4) "sixteenth-rest"
			  (if (= rhy 3/8) "dotted-sixteenth-rest"
			    (if (= rhy 1/8) "thirty-second-rest"
			      (if (= rhy 1/16) "sixty-fourth-rest"
				(format nil "rest (rq ~A)" (fratify rhy))))))))))))))))))

(defmethod rest-p ((obj t)) nil)
(defmethod rest-p ((obj rest-mixin)) t)

(defmethod descry ((rest rest-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A~A~A~A~A~A"
	  (if (not controller) "(rest" "")
	  (if (rest-mark rest) (format nil " :mark ~A" (rest-mark rest)) "")
	  (if (rest-line rest) (format nil " :line ~A" (rest-line rest)) "")
	  (if (rest-pause rest) (format nil " :pause ~A" (rest-pause rest)) "")
	  (if (ties rest) (format nil " :ties (list ~{~A ~})" (loop for tie in (ties rest) collect (descry tie))) "")
	  (if (tie-direction rest) (format nil " :tie-direction :~(~A~)" (tie-direction rest)) "")
	  (if (beamed rest) " :beamed t" "")
	  (if (outer-beam rest)
	      (format nil " :outer-beam '(~{~1,3F ~})" (outer-beam rest))
	    "")
	  (if (beams rest) (format nil " :beams ~A~%                    " (descry (beams rest))) "")
	  (if (store-data rest) (format nil " :store-data ~A" (store-data rest)) "")
	  (if (next-method-p) (call-next-method rest stream (or controller rest)) "")
	  (if (not controller) ")" "")))

(defmethod identify ((rst rest-mixin))
  (format nil "(~A~A~A~A~A~A~A)"
	  (rest-name rst)
 	  (if (onset rst) (format nil " (onset ~A)" (fratify (onset rst))) "")
	  (if (rest-pause rst) (format nil " ~A" (identify (rest-pause rst))) "")
	  (if (ties rst) (format nil "~{ ~(~A~)~}" (sorted-ties (ties rst))) "")
	  (if (tie-direction rst) (format nil "(tie-direction :~(~A~))" (tie-direction rst)) "")
	  (if (store-data rst) (identify-store-data rst) "")
	  (the-usual-suspects rst)))

(defmethod copy ((old-rest rest-mixin) &optional object)
  (let ((new-rest (if (not object) (make-simple-rest-1)
		    (if (write-protected object) (copy object)
		      object))))
    (if (rest-pause old-rest) (setf (rest-pause new-rest) (copy (rest-pause old-rest))))
    (setf (rest-mark new-rest) (rest-mark old-rest))
    (setf (rest-line new-rest) (rest-line old-rest))
    (setf (beamed new-rest) (beamed old-rest))
    (setf (outer-beam new-rest) (outer-beam old-rest))
    (if (beams old-rest) (setf (beams new-rest) (copy (beams old-rest))))
    (if (ties old-rest) (setf (ties new-rest) (loop for tie in (ties old-rest) collect (copy tie))))
    (setf (tie-direction new-rest) (tie-direction old-rest))
    (if (next-method-p) (call-next-method old-rest new-rest))
    new-rest))

(defmethod setf-rhythm ((rest rest-mixin) ur-val)
  (let ((val (fratify ur-val)))
    (setf (rest-mark rest) (copy (get-rest-mark val)))
    (setf (odb-duration rest) val)
    (setf (dots rest) (how-many-dots val))
    (setf (flags rest) (how-many-flags val))
    (setf (quarters rest) val)))

(defmethod get-tremolo ((rest rest-mixin)) nil)
(defmethod (setf stem-direction) (val (rest rest-mixin)) (declare (ignore val)))
(defmethod stem-dy ((rest rest-mixin)) 0)
(defmethod stem-mark ((rest rest-mixin)) nil)
(defmethod collision ((rest rest-mixin)) 0)
(defmethod (setf collision) (val (rest rest-mixin)) (declare (ignore val)))

(defmethod display ((rest rest-mixin) container score &rest others)
  (declare (ignore container))
  ;; line is the clef-relative center line of the rest (for dot placement)
  ;; the Sonata font treats x0 y0 of the rest as the point of the lowest flag, so we need to fix up everything
  (when (not (invisible-matrix-p rest))
    (let* ((x0 (if (or (not (member (rest-mark rest) (list %wholerest %measurerest %repsign)))
		       (member :unmetered (store-data rest)))
		   (+ (box-x0 rest) (dxy-dx rest))
		 (- (+ (* .5 (+ (box-x0 rest) (box-x1 rest))) (dxy-dx rest)) (center rest))))
	   (line (if (eq (rest-mark rest) %repsign) 2 (or (rest-line rest) 4)))
	   (y0 (+ (%staff-y0 rest) (dxy-dy rest) (* line (staff-line-separation score))))
	   (rm (rest-mark rest))
	   (dots (dots rest))
	   (flags (flags rest))
	   (ty0 (+ (if (zerop flags)
		       y0
		     (if (= flags 1)
			 (+ y0 (staff-line-separation score))
		       (if (evenp flags)
			   (- y0 (* (1- flags) (staff-line-separation score)))
			 (if (= flags 3)
			     (- y0 (staff-line-separation score))
			   (- y0 (* flags (staff-line-separation score)))))))
		   (if (oddp line) (staff-line-separation score) 0)))) ;always position in a space, not on a line

      #+Petrucci-Medium (if (= flags 1) (decf ty0 .125) (if (> flags 1) (incf ty0 .125) (if (eq (rest-mark rest) %halfrest) (incf ty0 .01))))
      
      (setf  (box-y0 rest) (+ (%staff-y0 rest) (* line (staff-line-separation score))))
      
      (when (rest-pause rest) (display (rest-pause rest) rest score))

      (if (and (beams rest) 
	       (not (member :just-fooling others))
	       (not (member :just-do-it others)))
	  (display (beams rest) rest score))

      (if (marks rest) (apply #'display-marks rest score others))
      
      (moveto score x0 ty0)
      (if (not (identity-matrix-p rest))
	  (show score (glyph rm (matrix (matrix rest))))
	(show score rm))
      (let ((mn (and (store-data rest) 
		     (find-if #'(lambda (n) (and (listp n) (eq (first n) :measure-number))) (store-data rest)))))
	(when mn
	  (let ((num (second mn)))
	    (moveto score (- x0 (if (> num 10) (if (> num 100) .3 .15) 0)) (+ y0 1.2))
	    (show score (number-to-glyph num)))))
      (when (and dots (plusp dots)
		 (not (member rest '(whole-rest measure-rest one-twenty-eighth-rest))))
	(let* ((dx (+ (dot-distance score) (max .2 (g-rx rm))))
	       (dy (if (oddp line)
		       0
		     (staff-line-separation score) )))
	  (moveto score (+ x0 dx) (+ y0 dy))
	  (if (= dots 1)
	      (simple-show score %dot)
	    (show score (apply #'glyphs (loop for dot from 0 below dots 
					 collect %dot
					 if (< dot (1- dots)) collect %space)))))))))

(defmethod notify ((rest rest-mixin) &optional objects)
  (apply #'rest-1 rest objects))

(defun rest-1 (&rest objects)		;for backwards compatibility
  (apply #'ur-rest (make-simple-rest-1) objects))

(defun rest (&rest objects)
  (apply #'ur-rest (make-simple-rest-1) objects))

(defun ur-rest (new-rest &rest objects)
  (loop for act in objects do
    (if (self-acting-p act)
	(funcall (action act) new-rest (arguments act))
      (if (rest-p act)
	  (copy act new-rest)

	(if (rhythm-p act)
	    (progn
	      (rcopy act new-rest)
	      (setf (rest-mark new-rest) (copy (get-rest-mark (quarters new-rest)))))

	  (if (pause-p act)
	      (setf (rest-pause new-rest) (copy act))
	    (if (or (text-p act) (sundry-p act) (glyph-list-p act))
		(add-to-marks new-rest (list act))
	      (if (numberp act)
		  (push (list :measure-number act) (store-data new-rest))
		(cmn-warn "odd argument to rest: ~A" act))))))))
  new-rest)

(defparameter rest-walls '(.05 0 .05 0))
(defparameter rest-fences '(.1 0 .1 0))
(defparameter rest-expanders '(2 0 2 0))

(defmethod house ((rest rest-mixin) score)
  (when (not (eq (visible-justification rest) :none))
    (setf (box-x1 rest) (+ (max .2 (g-rx (rest-mark rest)))
			   (if (and (dots rest) (plusp (dots rest)))
			       (+ (dot-distance score)
				  (* (dots rest) (g-rx %dot))
				  (* (1- (dots rest)) (g-rx %space)))
			     0)))
    (setf (center rest) (* .5 (max .2 (g-rx (rest-mark rest))))) ;i.e. ignore dots when aligning vertically
    (setf (walls rest) rest-walls)
    (setf (fences rest) rest-fences)
    (if (not (expanders rest))
	(setf (expanders rest) rest-expanders))))

(defun make-rest (dur)
  (if (>= dur 8) (copy double-whole-rest)
    (if (>= dur 4) (copy whole-rest)
      (if (>= dur 3) (copy dotted-half-rest)
	(if (>= dur 2) (copy half-rest)
	  (if (>= dur 3/2) (copy dotted-quarter-rest)
	    (if (>= dur 1) (copy quarter-rest)
	      (if (>= dur 3/4) (copy dotted-eighth-rest)
		(if (>= dur 1/2) (copy eighth-rest)
		  (if (>= dur 1/4) (copy sixteenth-rest)
		    (copy thirty-second-rest)))))))))))

(defun get-rest-mark (dur)
  (if (>= dur 8) (rest-mark double-whole-rest)
    (if (>= dur 4) (rest-mark whole-rest)
      (if (>= dur 3) (rest-mark dotted-half-rest)
	(if (>= dur 2) (rest-mark half-rest)
	  (if (>= dur 3/2) (rest-mark dotted-quarter-rest)
	    (if (>= dur 1) (rest-mark quarter-rest)
	      (if (>= dur 3/4) (rest-mark dotted-eighth-rest)
		(if (>= dur 1/2) (rest-mark eighth-rest)
		  (if (>= dur 1/4) (rest-mark sixteenth-rest)
		    (rest-mark thirty-second-rest)))))))))))




;;;
;;; ----------------   grace notes
;;;
;;; these are normal notes with small heads, beams, stems, sometimes with a slash
;;; they have no duration, but take up lots of horizontal space, so they are
;;; implemented here as marks -- in this case the mark (i.e. sundry) contains
;;; notes to be displayed.  stem-direction defaults to :up. Slashed if one note.
;;; slurred is default.
;;;
;;; (c4 e (grace-note d4)) (c4 e (grace-notes d4 fs4 e4)) (c4 e (grace-notes (slashed nil) (stem-direction :down) a4 f4))
;;; similarly, a chord can have grace-notes: (chord (notes c4 e4 g4) q (grace-notes (chord d4 f4 a4)))

(defclass grace-note (score-object)
  ((data :initarg :data :initform nil :accessor grace-data) ;a list of notes and chords, like chord-data list
   (stem-direction :initarg :stem-direction :initform :up :accessor stem-direction)
   ;; has to be :up, not :up? -- otherwise gets upended at dumb times
   (slashed :initarg :slashed :initform t :accessor slashed) ;t if one note, nil otherwise is default
   (slurred :initarg :slurred :initform t :accessor slurred) ;if nil, no slur to main note
   (user-data :initarg :user-data :initform nil :accessor slur-user-data)
   (direction :initarg :direction :initform nil :accessor slur-direction)
   (thickness :initarg :thickness :initform nil :accessor slur-thickness)
   (curvature :initarg :curvature :initform nil :accessor slur-curvature)))

(defmethod grace-note-p ((obj t)) nil)
(defmethod grace-note-p ((obj grace-note)) t)

(self-action slashed setf-slashed)
(self-action slurred setf-slurred)

(defun grace-note (&rest objects)
  (let ((new-grace (make-grace-note))
	(nlen 0)
	(bt 0.0))
    (let ((new-staff (make-staff)))
      ;; let the args to grace-notes mimic the staff style input (for appogiaturas primarily)
      (multiple-value-bind
	  (data actions local-brace)
	  (notify new-staff objects)
	(declare (ignore local-brace))
	(loop for act in actions do
	  (funcall (action act) new-grace (arguments act)))
	(setf (grace-data new-grace) data)))
    
    (setf nlen (length (grace-data new-grace)))
    (loop for note in (grace-data new-grace) do
      (if (not (audible-stem-direction note)) 
	  (setf (audible-stem-direction note) (stem-direction new-grace)))
      (if (or (not (flags note)) 
	      (zerop (flags note)))
	  (if (= nlen 1) (setf (flags note) 1)
	    (if (< nlen 4) (setf (flags note) 2)
	      (setf (flags note) 3))))
      (if (not (quarters note)) 
	  (setf (quarters note) 
	    (if (= nlen 1) 1/2
	      (if (= nlen 2) 1/4
		1/8))))
      (setf (odb-beat note) bt)
      (incf bt (quarters note)))
    (make-self-acting :action #'add-to-marks :arguments (list new-grace))))

(defun grace-notes (&rest objects) (apply #'grace-note objects))
(defun appogiatura (&rest objects) (apply #'grace-note (slashed nil) objects))

(defmethod descry ((grace grace-note) &optional stream controller)
  (format stream "(grace-note :slashed ~(~A~) :slurred ~(~A~) :stem-direction :~(~A~)~A :data (list~{ ~A~}))"
	  (slashed grace) (slurred grace) (stem-direction grace)
	  (if (next-method-p) (call-next-method grace stream (or controller grace)) "")
	  (loop for note in (grace-data grace) collect (descry note stream grace))))

(defmethod identify ((gn grace-note))
  (format nil " (~A~A~A~A~{ ~A~})"
	  (if (not (slashed gn)) "appogiatura" "grace-note")
	  (if (not (slurred gn)) " (slurred nil)" "")
	  (if (not (eq (stem-direction gn) :up)) " stem-down" "")
	  (if (eq (visible-justification gn) :none) " unjustified" "")
	  (loop for note in (grace-data gn) 
	   collect (if (or (marks note)
			   (/= (quarters note) 1))
		       (identify note)
		     (note-name note)))))

(defmethod copy ((grace grace-note) &optional object)
  (let ((new-grace (if (not object) (make-grace-note)
		     (if (write-protected object) (copy object)
		       object))))
    (setf (slashed new-grace) (slashed grace))
    (setf (slurred new-grace) (slurred grace))
    (setf (stem-direction new-grace) (stem-direction grace))
    (setf (grace-data new-grace) (loop for note in (grace-data grace) collect (copy note)))
    (if (next-method-p) (call-next-method grace new-grace))
    new-grace))

(defmethod display ((grace grace-note) outer-note score &rest rest)
  (let* ((grace-scale (grace-note-size score))
	 (inverse-scale (/ 1.0 grace-scale))
	 (old-separation (staff-line-separation score))
	 (glen (length (grace-data grace)))
	 (outer-line (minimum-line outer-note))
	 (grace-dxs (loop for note in (grace-data grace) 
		     collect (if (note-p note)
				 (+ .5 
				    (if (sign note) (+ (g-rx (sign note)) (accidental-breathing-space score)) 0)
				    (if (or (< (line note) -1) 
					    (> (line note) 9))
					.1
				      0))
			       (+ .5 
				  (if (find-if #'sign (chord-data note))
				      (let ((all-sign-lines (sort (loop for cdn in (chord-data note) 
								   if (note-sign cdn) collect (line cdn)) 
								  #'>)))
					(if (and (> (length all-sign-lines) 1)
						 (< (loop for n1 in all-sign-lines and n2 in (cdr all-sign-lines) 
						     minimize (abs (- n2 n1))) 
						    3))
					    1.0
					  .5))
				    0)))))
	 (x-space (+ .1 (* grace-scale (apply #'+ grace-dxs)))))
    (if (member :just-fooling rest)
	(if (not (eq (visible-justification grace) :none))
	    (moveto score (- (box-x0 outer-note) x-space) (box-y0 outer-note)))
      (progn
	(setf (staff-line-separation score) (* inverse-scale old-separation))
	(with-scaling score 
		      grace-scale 
		      (* (- (+ (box-x0 outer-note) (dxy-dx grace)) x-space) (scr-size score))
		      (* (%staff-y0 outer-note) (scr-size score))
		      (let ((fc nil)
			    (xorig 0.0))
			(loop for note in (grace-data grace) and dx in grace-dxs do
			  (setf (%staff-y0 note) 0.0)
			  (setf (box-y0 note) (* (minimum-line note) (staff-line-separation score)))
			  
			  (if (or (note-p note)
				  (not (find-if #'sign (chord-data note))))
			      (setf (center note) (+ .15 
						     (if (and (sign note) (zerop (audible-collision note)))
							 (+ (g-rx (sign note))
							    (if (or (<= (minimum-line note) -2) (<= 10 (maximum-line note)))
								(leger-line-length score)
							      0)
							    (accidental-breathing-space score))
						       0)))
			    (let ((old-output (output score)))
			      ;; chords with accidentals can be all over the map, so we have to draw them to figure out where the stem is
			      (clear-server score)
			      (setf (center note) (display note nil score :just-fooling))
			      (setf (output score) old-output)))
			  
			  (setf (box-x0 note) xorig)
			  (incf xorig dx))
			
			(if (> glen 1)
			  (annotate-beam score nil 0 
					 (first (grace-data grace)) 
					 (grace-data grace) nil 
					 (stem-direction grace) 
					 (* grace-scale (staff-line-separation score))))
			;; beam-unit happens in display
			
			(if (slurred grace)
			    (if (> glen 1)
				(prepare-slur score 
					      (make-slur :type :left 
							 :curvature (slur-curvature grace) 
							 :thickness (slur-thickness grace))
					      (append (grace-data grace) 
						      (list (make-note :stem-direction :up?
								       :y0 (- (* outer-line (staff-line-separation score)) (/ .15 grace-scale))
								       :staff-y0 0
								       :x0 (+ xorig (/ (if (stem-is-up? outer-note) .2 0) grace-scale))
								       :line (if (stem-is-down? outer-note) outer-line (1- outer-line))
								       :sign (and (sign outer-note) (copy (sign outer-note)))
								       :note-head (note-head outer-note)))))
			      ;; two note "slur" is so problematic that I've decided (after n failures)
			      ;;   to handle them directly -- no intervening two-note-slur calls and so on
			      (let* ((gn (first (grace-data grace))) ;the actual grace note
				     (on outer-note)
				     (gn-stem (or (stem-direction gn) (stem-direction grace) :up?))
				     (gn-stem-up (member gn-stem '(:up :up?)))
				     (on-stem (or (stem-direction outer-note)
						  (and (< (head-line outer-note) 4) :up?)
						  :down?))
				     (on-stem-up (member on-stem '(:up :up?)))
				     (stems-agree (or (and gn-stem-up on-stem-up)
						      (and (not gn-stem-up) (not on-stem-up))))
				     (gn-stem-end (+ (get-stem-end score gn (member gn-stem '(:up :up?)) '(:grace-note))
						     (or (stem-dy gn) 0)))
				     (gn-line (if (not gn-stem-up) (maximum-line gn) (minimum-line gn)))
				     (on-line (if (not on-stem-up) (maximum-line on) (minimum-line on)))
				     (line-diff (- gn-line on-line)) ;true (unscaled) line diff
				     (close-heads (< (abs line-diff) 4))
				     (slur-dir (or (slur-direction grace)
						   (if (or stems-agree close-heads)
						       (if gn-stem-up 
							   :down 
							 :up)
						     ;; here we have the weird cases -- opposite stems, wide leaps
						     (if (or (minusp line-diff)
							     (> gn-line 9))
							 :up
						       :down))))
				     (connection-position (if (eq slur-dir :up)
							      (if (and (minusp line-diff) gn-stem-up)
								  :stem-head
								:head-head)
							    (if (and (plusp line-diff) (not gn-stem-up))
								:stem-head
							      :head-head)))
				     (sudat (slur-user-data grace))
				     (sux0 (or (and sudat (first sudat)) 0))
				     (suy0 (or (and sudat (second sudat)) 0))
				     (sux1 (or (and sudat (third sudat)) 0))
				     (suy1 (or (and sudat (fourth sudat)) 0))
				     (sux2 (or (and sudat (fifth sudat)) 0))
				     (suy2 (or (and sudat (sixth sudat)) 0))
				     (sux3 (or (and sudat (seventh sudat)) 0))
				     (suy3 (or (and sudat (eighth sudat)) 0))

				     (gny0 (+ suy0 
					      (if (eq connection-position :head-head) 
						  (if (note-p gn) 
						      (y0 gn) 
						    (* (staff-line-separation score) gn-line))
						gn-stem-end)
					      (if (eq slur-dir :up) .2 -.3)
					      (if (and (< gn-line -2) (eq slur-dir :up))
						  (* (- (+ 2 gn-line)) (staff-line-separation score))
						0)
					      (if (and (> gn-line 10) (eq slur-dir :down))
						  (* (- 10 gn-line) (staff-line-separation score))
						0)
					      (if (and stems-agree
						       (or (and gn-stem-up (< line-diff -4))
							   (and (not gn-stem-up) (> line-diff 4))))
						  (if gn-stem-up .2 -.2)
						0)))

				     (gnxscl (/ .5 (grace-note-size score)))

				     (gny3 (+ suy3 
					      (* on-line (staff-line-separation score))
					      (* gnxscl
						 (if (eq slur-dir :down)
						     (if (and (not close-heads)
							      (plusp line-diff)
							      (or (eq connection-position :head-head)
								  (> line-diff 6)))
							 .1
						       (if (sign on)
							   -.8
							 -.4))
						   (if (and (not close-heads)
							    (minusp line-diff)
							    (or (eq connection-position :head-head)
								(< line-diff -6)))
						       .1
						     (if (sign on)
							 .8
						       .4))))))

				     (gnx0 (+ sux0 (box-x0 gn) (center gn) 
					      (if (and (chord-p gn) (sign gn)) .25 0)
					      (if (or (and gn-stem-up 
							   (eq slur-dir :up) 
							   (plusp line-diff)) 
						      (and stems-agree
							   (or (and gn-stem-up (< line-diff -4))
							       (and (not gn-stem-up) (> line-diff 4)))))
						  (+ .3 (if (or (> gn-line 10) (< gn-line -2)) .1 0))
						0)))
				     (gnx3 (+ sux3 xorig 
					      (* gnxscl
						 (if (or (and (or (> line-diff 6)
								  (and (> line-diff 3)
								       (eq connection-position :head-head)))
							      (eq slur-dir :down))
							 (and (or (< line-diff -6)
								  (and (< line-diff -3)
								       (eq connection-position :head-head)))
							      (eq slur-dir :up))
							 (and (eq slur-dir :down)
							      (not on-stem-up)))
						     .10
						   (if (sign on)
						       1.0
						     .4)))))
				     ;; now take into account curvature, thickness, and user-data, and general "nice" shapes
				     (dx (- gnx3 gnx0))
				     (dy (- gny3 gny0))
				     (len (sqrt (+ (* dx dx) (* dy dy))))
				     (gnslope (divide dy dx))
				     (gnx1 (+ gnx0 sux1 (* .33 dx)))
				     (gnx2 (+ gnx0 sux2 (* .67 dx)))
				     (curve (* (or (slur-curvature grace)
						   (slur-curvature score))
					       (if (< len .7) .5 1.0)))
				     (curve-fixup (if (> (abs gnslope) 6)
						      1.0
						    (if (> (abs gnslope) 3)
							0.5
						      0)))
				     (bow-curve (if on-stem-up
						    (if gn-stem-up
							(- curve)
						      (if (eq slur-dir :up)
							  curve
							(- curve)))
						  (if (not gn-stem-up)
						      curve
						    (if (eq slur-dir :down)
							(- curve)
						      curve))))
				     (bow-curve-fixup (if (minusp bow-curve) (- curve-fixup) curve-fixup))
				     (thick (or (slur-thickness grace)
						(slur-thickness score)))
				     (gny1 (+ gny0 suy1 (* (- gnx1 gnx0) gnslope) bow-curve bow-curve-fixup))
				     (gny2 (+ gny0 suy2 (* (- gnx2 gnx0) gnslope) bow-curve)))

				(moveto score gnx0 gny0)
				(curveto score gnx1 gny1 gnx2 gny2 gnx3 gny3)
				(curveto score gnx2 (- gny2 thick) gnx1 (- gny1 thick) gnx0 gny0)
				(fill-in score))))

			(loop for note in (grace-data grace) do
			  (if (not fc)	;first note flag (for slash mark)
			      (progn
				(setf fc (/ (display note nil score :grace-note) (if (chord-p note) 1.0 (scr-size score))))
				(when (slashed grace)
				  (if (= glen 1)
				      (progn
					(moveto score 
						(+ fc (if (stem-is-up? note) .05 -.2) (if (chord-p note) -.15 0))
						(if (and (note-p note)
							 (<= 0 (minimum-line note) (maximum-line note) 9))
						    (+ (box-y0 note)
						       (if (stem-is-up? note) .35 -.4))
						  (if (stem-is-up? note)
						      (- (or (stem-end note) 0)
							 .7)
						    (+ (or (stem-end note) 0) 
						       .7))))
					(setf (line-width score) (grace-note-slash-thickness score))
					(rlineto score 
						 (+ .6 (if (> (flags note) 2) .1 0))
						 (* (if (stem-is-up? note) 1.0 -1.0)
						    (+ .6 
						       (if (> (flags note) 2) 
							   (max 0 (- (stem-end note) (box-y0 note) 1.25))
							 0))))
					(simple-draw score)
					(setf (line-width score) 0))
				    (let* ((beam-data (outer-beam note))
					   ;; draw slash at front going up if stem up and beam flat or down
					   ;;            at front going down if stem down and beam up
					   ;;            at back going up if stem down and beam flat or down
					   ;;            at back going down if stem up and beam up
					   ;; we assume we've already called beamify which has set up outer-beam
					   (bx0 (first beam-data))
					   (by0 (second beam-data))
					   (bx1 (third beam-data))
					   (by1 (fourth beam-data))
					   (beam-up (> by1 by0))
					   (gstem-up (stem-is-up? note))
					   (slash-at-front (or (and gstem-up (not beam-up)) (and (not gstem-up) beam-up)))
					   (slash-going-up (or (and gstem-up (not beam-up)) (and (not gstem-up) (not beam-up))))
					   (beams (flags note))
					   (sx0 (if slash-at-front 
						    (- bx0 .25) 
						  (if (and slash-going-up (> beams 2))
						      (- bx1 .75)
						    (- bx1 .5))))
					   (bspace (* (1+ beams) (beam-spacing score)))
					   (dist (+ .2 (* bspace (if (or slash-going-up slash-at-front) 1.0 .75))))
					   (sx1 (+ sx0 dist))
					   (sy0 (if slash-going-up 
						    (if slash-at-front
							(- by0 -.2 bspace)
						      (- by1 .25))
						  (if slash-at-front
						      (+ by0 .25 bspace)
						    (+ by1 .3))))
					   (sy1 (if slash-going-up
						    (+ sy0 dist)
						  (- sy0 dist))))
				      (moveto score sx0 sy0)
				      (setf (line-width score) (grace-note-slash-thickness score))
				      (lineto score sx1 sy1)
				      (simple-draw score)
				      (setf (line-width score) 0)))))
			    (display note nil score)))))
	(setf (staff-line-separation score) old-separation)))))

#|
(cmn (size 40) (free-expansion-factor 1.0) staff treble 
     (g4 q (grace-note e4 e4)) 
     (g4 q (grace-note stem-down e4 e4))
     (g4 q (grace-note a4 e4))
     (g4 q (grace-note stem-down a4 e4))
     (g4 q (grace-note e4 a4))
     (g4 q (grace-note stem-down e4 a4))
     (line-break)
     (g4 q (grace-note e4 e e4 e)) 
     (g4 q (grace-note stem-down e4 e e4 e))
     (g4 q (grace-note a4 e e4 e))
     (g4 q (grace-note stem-down a4 e e4 e))
     (g4 q (grace-note e4 e a4 e))
     (g4 q (grace-note stem-down e4 e a4 e))
     (line-break)
     (g4 q (grace-note e4 f4 e4 f4)) 
     (g4 q (grace-note stem-down e4 f4 e4 b4))
     (g4 q (grace-note a4 f4 e4 c4))
     (g4 q (grace-note stem-down a4 f4 e4 c4))
     (g4 q (grace-note e4 f4 a4 b4))
     (line-break)
     (g4 q (grace-note stem-down e4 f4 a4 b4))
     (g4 q (grace-note (chord e4 b4))) 
     (g4 q (grace-note (chord ef4 bf4)))
     (g4 q (grace-note (chord ef4 af4 bf4 en5)))
     (g4 q (grace-note a4 (chord c4 c5)))
     (g4 q (grace-note (chord d4 a4) (chord c4 g4) (chord b3 fs4) (chord bf4 af4 gf4) (chord bf4 af4 ef4) (chord a3 e4)))
     (line-break)
     (chord (notes g4 d5) q (grace-note b4))
     (chord (notes gs4 df5) q (grace-note e4 f4 g4 a4))
     (g4 q (grace-note stem-down (chord e4 b4))) 
     (g4 q (grace-note stem-down (chord ef4 bf4))))

;;; see grace.cmn for many 2 note cases
(cmn (size 40) staff treble 
     (g4 q (grace-note fs4)) 
     (g4 q (grace-note gf4)) 
     (g4 q (grace-note ef4)) 
     (g4 q (grace-note b4)) 
     (b4 q (grace-note e5)) 
     (b4 q (grace-note d4)) 
     (g4 q (grace-note ef5)))

(cmn (size 40) staff treble 
     (g4 q (grace-note e4 fs4)) 
     (g4 q (grace-note gf4 ef4)) 
     (g4 q (grace-note ef4 c4)) 
     (g4 q (grace-note b4 a4)) 
     (b4 q (grace-note e5 d5 cs5)) 
     (b4 q (grace-note d4 e4 fs4 g4 a4)) 
     (g4 q (grace-note ef5 bf4 ef4)))

(cmn (size 60) staff treble a4 q. (grace-note f4 g4) g4 e f4 q g5 q. g5 e (grace-note a5) begin-slur f5 e g5 e end-slur)

(cmn (size 60) staff treble (c6 q (appogiatura b5 a5)) (c6 q (appogiatura bf5 as5)) (c5 q (appogiatura b4 d5)))
(cmn staff treble (c4 q (grace-note (notes d4 (chord fs4 bf4)))))
(cmn staff treble (c4 q (grace-note (notes d4 a4))))
|#



(defun auxiliary-note (&rest objects)
  (let ((nt (apply #'note objects)))
    (make-self-acting 
     :arguments nt
     :action #'(lambda (owner note) 
		 (add-to-marks 
		  owner
		  (list (make-sundry 
			 :name :auxiliary-note
			 :mark #'(lambda (mark outer-note score &rest rest)
				   (declare (ignore mark))
				   (if (member :note rest)
				       ;; this is for cmn-store's benefit
				       note
				     (let* ((grace-scale 
					     (if (and (matrix note) (not (identity-matrix-p note)))
						 (first (matrix note))
					       (grace-note-size score)))
					    (inverse-scale (/ 1.0 grace-scale))
					    (old-separation (staff-line-separation score)))
				       (if (not (member :just-fooling rest))
					   (progn
					     (setf (staff-line-separation score) (* inverse-scale old-separation))
					     (with-scaling score 
							   grace-scale 
							   (* (+ (box-x0 outer-note) (dxy-dx outer-note)) (scr-size score))
							   (* (%staff-y0 outer-note) (scr-size score))
					       (setf (staff-y0 note) 0.0)
					       (setf (line note) (place-of-note-given-note outer-note note))
					       (setf (box-y0 note) (* (line note) (staff-line-separation score)))
					       (setf (center note) (+ .15 
								      (if (and (sign note) (zerop (audible-collision note)))
									  (+ (g-rx (sign note))
									     (if (or (<= (minimum-line note) -2) (<= 10 (maximum-line note)))
										 (leger-line-length score)
									       0)
									     (accidental-breathing-space score))
									0)))
					       (setf (box-x0 note) 0)
					       (display note nil score))
					     (setf (staff-line-separation score) old-separation))
					 (moveto score (+ (box-x0 outer-note) (dxy-dx outer-note) (dxy-dx note)) 0))))))))))))



;;;
;;; ----------------   cycle
;;;

;;; also need a reader of note names that doesn't demand the explicit octave number
;;; (e f ff d g problems)

(defun cycle (&rest objects-or-lists)
  ;; take a sequence of sequences of cmn objects and parcel them out into a single full list (score-object-list)
  ;; these are expanded by notify staff during scorify
  (let* ((len (length objects-or-lists))
	 (args (make-array len))
	 (max-list 0)
	 (data nil)
	 (cur-args (make-array len)))
    (loop for i from 0 below len and object in objects-or-lists do
      (setf (aref args i) (if (listp object) object (list object)))
      (setf (aref cur-args i) (aref args i)))
    (setf max-list (loop for i from 0 below len maximize (length (aref args i))))
    (loop for i from 0 below max-list do
      (let ((owner (first (aref cur-args 0))))
	(push (funcall #'notify (if (write-protected owner)
				    (copy owner)
				  owner)
		       (loop for j from 1 below len collect (first (aref cur-args j)))) data))
      (loop for j from 0 below len do
	(setf (aref cur-args j) (cdr (aref cur-args j)))
	(if (not (aref cur-args j)) 
	    (setf (aref cur-args j) (aref args j)))))
    (engorge (nreverse data))))



;;;
;;; ----------------   add-note, display-notes
;;;
;;; clm/cm/stella interface -- each instrument gives us its name, begin time, duration, frequency, and maybe dynamics etc
;;; we turn out the equivalent cmn input stream (at level of notified staff data)
;;;
;;; examples:
;;;
;;; (display-notes nil (list (list 'hi 0 1 440 .1 staccato) '(hi 1 1 660 .1)))
;;; (display-notes (list (metronome 75) (size 24) (title "Opus 1")) (list (list 'hi 0 .2 440 .1 staccato) '(hi .2 .6 660 .1)))


(defun freq-to-pitch (freq)
  (floor (* 12 (+ (log (/ (if (numberp freq) freq (eval freq)) 16.351) 2.0) (/ 1.0 24)))))

(defvar last-clm-sign nil)

(defun pitch-to-note-octave-and-accidental (pitch) ;for clm note lists
  (let* ((pclass (mod pitch 12))
	 (octave (floor pitch 12))
	 (cclass (case pclass 
		   (0 0) 
		   (1 (if (not (eq last-clm-sign flat)) 0 1))
		   (2 1) 
		   (3 (if (not (eq last-clm-sign sharp)) 2 1))
		   (4 2) 
		   (5 3) 
		   (6 (if (not (eq last-clm-sign flat)) 3 4)) 
		   (7 4) 
		   (8 (if (not (eq last-clm-sign sharp)) 5 4))
		   (9 5) 
		   (10 (if (not (eq last-clm-sign sharp)) 6 5) )
		   (11 6))))
    (if (member pclass '(0 2 4 5 7 9 11))
	(list pclass octave (setf last-clm-sign nil) cclass)
      (if (or (and (= pclass 1) (= cclass 0))
	      (and (= pclass 3) (= cclass 1))
	      (and (= pclass 6) (= cclass 3))
	      (and (= pclass 8) (= cclass 4))
	      (and (= pclass 10) (= cclass 5)))
	  (list pclass octave (setf last-clm-sign sharp) cclass)
	(list pclass octave (setf last-clm-sign flat) cclass)))))

(defun freq-to-note-octave-and-accidental (freq)
  (pitch-to-note-octave-and-accidental (freq-to-pitch freq)))

(defun pitch-of (note)
  (+ (pitch note) (* 12 (octave note))))

(defun name-to-note-and-accidental (name)
  ;; give score-style note name (optional sign and octave), returns pitch, c-class, sign
  (let* ((string-name (format nil "~s" name))
	 (cclass-name (elt string-name 0))
	 (name-len (length string-name))
	 (csign (and (> name-len 1) 
		     (member (elt string-name 1) '(#\s #\s #\f #\f #\n #\n)) 
		     (elt string-name 1)))
	 (sign (if csign (if (member csign '(#\s #\s)) sharp 
			   (if (member csign '(#\f #\f)) flat
			     natural))
		 nil))
	 (octave (or (and (> name-len (if csign 2 1)) 
			  (char<= #\0 (elt string-name (if csign 2 1)) #\9)
			  (- (char-code (elt string-name (if csign 2 1))) 48))
		     old-octave))
	 (cclass (if (upper-case-p cclass-name)
		     (position cclass-name '(#\c #\d #\e #\f #\g #\a #\b))
		   (position cclass-name '(#\c #\d #\e #\f #\g #\a #\b))))
	 (pclass (+ (case cclass (0 0) (1 2) (2 4) (3 5) (4 7) (5 9) (6 11))
		    (if (and sign (eq sign sharp)) 1
		      (if (and sign (eq sign flat)) -1
			0)))))
    (setf old-octave octave)
    (list pclass octave sign cclass)))

(defun duration-to-rhythm (score dur)
  (let* ((qnotes (* dur (/  (metronome score) (* (beats-per-quarter-note score) 60))))
	 (ttnotes (floor (* qnotes 8))))
    (case ttnotes
      (1 thirty-second)
      (2 s)
      (3 s.)
      ((4 5) eighth)
      (6 e.)
      ((7 8 9 10 11) q)
      ((12 13 14) q.)
      ((15 16 17 18 19 20 21 22) h)
      ((23 24 25 26) h.)
      ((27 28 29) h..)
      (otherwise w))))


(defvar staff-aliases nil)
(defvar staff-descriptors nil)

(defun find-staff (score name)
  (when (not staff-aliases)
    (setf staff-aliases (make-hash-table :test #'equal))
    (push (make-system) (systems score)))
  (or (gethash name staff-aliases)
      (let ((stf (make-staff :name name)))
	(setf (gethash name staff-aliases) stf)
	(push stf (staves (first (systems score))))
	stf)))

(defun add-staff (score staff-name other-names &rest acts)
  (let ((stf (make-staff :name staff-name)))
    (if (not staff-aliases) (setf staff-aliases (make-hash-table :test #'equal)))
    (setf (gethash staff-name staff-aliases) stf)
    (loop for name in other-names do
      (setf (gethash name staff-aliases) stf))
    (loop for act in acts do
      (if (self-acting-p act)
	  (funcall (action act) stf (arguments act))))
    (if (not (systems score)) (push (make-system) (systems score)))
    (push stf (staves (first (systems score))))
    stf))

(defvar accumulated-tempo-offset 0.0)
(defvar last-tempo-true-begin-time 0.0)
;;; these help handle tempo changes while collecting score data.
;;; the first is the integrated cmn-time at the point of the last tempo change.
;;; the second is the user's idea of the corresponding musical time of that tempo change.

(defun add-note (score name begin-time duration frequency &rest rest)
  (apply #' add-note-to-staff score (find-staff score name) begin-time duration frequency rest))

(defun add-note-to-staff (score staff begin-time duration frequency &rest rest)
  (let* ((pitch-data (if (numberp frequency)
			 (freq-to-note-octave-and-accidental frequency)
		       (list (pitch frequency) (octave frequency) (sign frequency) (cclass frequency))))
	 (nt (make-note :onset (+ accumulated-tempo-offset 
				  (* .25 (round (* 4 (* (/ (metronome score) 60) 
							(- begin-time last-tempo-true-begin-time))))))
			:octave (second pitch-data)
			:pitch (first pitch-data)
			:sign (third pitch-data)
			:cclass (fourth pitch-data))))
    (setf-rhythm nt (max .25 (* .25 (round (* 4 (/ (metronome score) 60) duration)))))
    (if rest
	(let ((with-errors nil))
	  (push (apply #'note nt rest) (staff-data staff)))
      (push nt (staff-data staff)))))

(defun add-data (score name obj)
  (add-data-1 score (find-staff score name) obj))

(defun add-data-1 (score staff obj)
  (declare (ignore score))
  (if (self-acting-p obj)
      (funcall (action obj) 
               (or (find-if #'note-p (staff-data staff))
                   (error "Can't find note for ~S in ~S" obj staff))
               (arguments obj))
    (if (or (sundry-p obj) (dynamics-p obj))
        (add-to-marks (or (and (staff-data staff) 
                               (first (staff-data staff)))
                          staff)
                      (list (if (write-protected obj) 
				(copy obj) obj)))
      (if (write-protected obj)
          (push (copy obj) (staff-data staff))
        (push obj (staff-data staff))))))

;; here if the datum calls the onset method (as in (bar (onset ...))), there should
;; first be some sort of fixup done on the onset value.


(defun set-cmn-tempo (score onset new-tempo)
  (let ((apparent-onset (+ accumulated-tempo-offset 
			   (* .25 (round (* 4 (* (/ (metronome score) 60) 
						 (- onset last-tempo-true-begin-time))))))))
    (setf accumulated-tempo-offset apparent-onset)
    (setf last-tempo-true-begin-time onset)
    (setf (metronome score) new-tempo)
    new-tempo))

(defun init-clm-input (&optional score) 
  (setf last-clm-sign nil)
  (setf staff-aliases nil)
  (setf staff-descriptors nil)
  (setf accumulated-tempo-offset 0.0)
  (setf last-tempo-true-begin-time 0.0)
  (setf *cmn-score* (if score
                        (reinitialize-score score)
                      (make-score))))

(defun reinitialize-score (score)
  ;; reset internal slots and return score for reuse. hope i got all of them.
  (dolist (slot '(staves pages lines line-data time-lines time-line systems))
    (setf (slot-value score slot) nil))
  score)


(defun begin-delayed-slur () (make-sundry :name :start-slur))
(defvar begin-delayed-slur (make-instance 'write-protected-sundry :name :start-slur))
(defun end-delayed-slur () (make-sundry :name :finish-slur))
(defvar end-delayed-slur (make-instance 'write-protected-sundry :name :finish-slur))

(defun deal-with-delayed-slurs (obj cnote)
  (if (marks cnote)
      (let ((ss (find :start-slur (marks cnote) :key #'sundry-name))
	    (fs (find :finish-slur (marks cnote) :key #'sundry-name)))
	(when ss
	  (user-add-to-left-slurs obj nil)
	  (setf (marks cnote) (remove ss (marks cnote))))
	(when fs
	  (user-add-to-right-slurs obj nil)
	  (setf (marks cnote) (remove fs (marks cnote)))))))

(defun check-for-delayed-slurs (obj)
  (when obj
    (if (chord-p obj)
	(loop for cnote in (chord-data obj) do
	  (deal-with-delayed-slurs obj cnote))
      (deal-with-delayed-slurs obj obj)))
  obj)
		       

(defun check-for-chords (objects)
  (let* ((new-data nil)
	 (current-chord nil)
	 (last-object nil)
	 (last-onset nil)
	 (robjects (reverse objects)))
    (loop for object in robjects do
      (let ((this-onset (odb-onset object)))
	(if (and (audible-p object)
		 last-onset
		 (= last-onset this-onset))
	    (if current-chord
		(push object (chord-data current-chord))
	      (if (audible-p last-object)
		  (setf current-chord (chord (notes last-object object) (onset this-onset) (duration (odb-duration object))))
		(progn
		  (push last-object new-data)
		  (setf last-object object))))
	  (progn
	    (if current-chord
		(progn
		  (push (check-for-delayed-slurs current-chord) new-data)
		  (setf current-chord nil))
	      (if last-object
		  (push (check-for-delayed-slurs last-object) new-data)))
	    (setf last-object object)
	    (setf last-onset this-onset)))))
    (push (check-for-delayed-slurs (or current-chord last-object)) new-data)
    (nreverse new-data)))

(defstruct stfdat top-freq bottom-freq above-g3 below-g4 mean-freq staff tstf bstf uclf uclfs ustfs)

(defun set-staff-clef (staff main-clefs &rest other-clefs)
  (let ((stfd (find staff staff-descriptors :key #'stfdat-staff)))
    (setf (stfdat-uclf stfd) main-clefs)
    (setf (stfdat-uclfs stfd) other-clefs)))

(defun set-staff-number (staff number)
  (let ((stfd (find staff staff-descriptors :key #'stfdat-staff)))
    (setf (stfdat-ustfs stfd) number)))

(defun finish-clm-input (score output-type with-onsets &rest descriptors)
  ;; until we get a better idea, descriptors is a list of lists,
  ;; each list being (function-name staff-name &rest other args)
  ;; we will run through that list, matching the staff-name to the appropriate staff,
  ;; then applying the first name to the resultant arg list.
  (let ((new-staves nil)
	(which-clefs nil))
    (setf staff-descriptors nil)
    (map-over-staves #'(lambda (score stf)
			 (declare (ignore score))
			 (let ((topf 0) 
			       (botf 100000)
			       (sumf 0)
			       (ctf 0)
			       (ag3 0)
			       (bg4 0))
			   (loop for object in (staff-data stf) do
			     (when (audible-p object)
			       (let ((pcl (+ (pitch object) (* 12 (octave object)))))
				 (incf sumf pcl)
				 (incf ctf)
				 (setf topf (max topf pcl))
				 (setf botf (min botf pcl))
				 (if (> pcl 43) (incf ag3))
				 (if (< pcl 55) (incf bg4)))))
			   (push (make-stfdat :top-freq topf :bottom-freq botf :mean-freq (divide sumf ctf)
					      :above-g3 ag3 :below-g4 bg4 :staff stf)
				 staff-descriptors)))
		     score)

    (setf staff-descriptors (sort staff-descriptors #'> :key #'stfdat-mean-freq))
    (loop for desc in descriptors do
      (let ((func (first desc))
	    (args (cddr desc))
	    (staff (find-staff score (second desc))))
	(apply func staff args)))

    (loop for stf in staff-descriptors do
      ;; here check first for any user-set decisions about number of staves and respective clefs
      (if (stfdat-ustfs stf) 
	  (if (= (stfdat-ustfs stf) 0) (setf which-clefs :neither)
	    (if (= (stfdat-ustfs stf) 2) (setf which-clefs :both)
	      (if (= (stfdat-ustfs stf) 1) 
		  (if (stfdat-uclf stf) (setf which-clefs (stfdat-uclf stf))))))) ;:treble for example
      (if (and (not which-clefs)
	       (stfdat-uclf stf) 
	       (not (stfdat-uclfs stf)))
	  (setf which-clefs (stfdat-uclf stf)))
      (if (not which-clefs)		;i.e. either not ustfs or uclfs+ustfs leave us undecided so far
	  (if (and (or (> (stfdat-bottom-freq stf) 48)
		       (> (stfdat-top-freq stf) 57)
		       (> (stfdat-above-g3 stf) 4))
		   (or (not (stfdat-uclfs stf))
		       (member :treble (stfdat-uclfs stf))))
	      (setf which-clefs :treble?)))
      (if (not which-clefs)
	  (if (stfdat-uclf stf) (setf which-clefs (stfdat-uclf stf))
	    (if (stfdat-uclfs stf)
		(if (member :bass (stfdat-uclfs stf)) (setf which-clefs :bass)))))
      (if (and (eq which-clefs :treble?)
	       (not (stfdat-ustfs stf)))
	  (if (or (< (stfdat-top-freq stf) 48)
		  (< (stfdat-bottom-freq stf) 40))
	      (setf which-clefs :both)))
      (if (eq which-clefs :treble?) (setf which-clefs :treble))
      (if (not which-clefs) (setf which-clefs :treble))
      
      (if (or (member which-clefs '(:treble :both))
	      (not (member which-clefs '(:bass :neither))))
	  (progn
	    (setf (stfdat-tstf stf) (make-staff :name (staff-name (stfdat-staff stf))))
	    (add-to-marks (stfdat-tstf stf) (list (make-staff-name)))
	    (push (stfdat-tstf stf) new-staves)))
      (if (member which-clefs '(:bass :both))
	  (progn
	    (setf (stfdat-bstf stf) (make-staff :name (staff-name (stfdat-staff stf))))
	    (if (stfdat-tstf stf) (setf (staff-local-brace (stfdat-bstf stf)) brace))
	    (push (stfdat-bstf stf) new-staves)))

      (if (marks (stfdat-staff stf)) 
	  (add-to-marks (or (stfdat-tstf stf) (stfdat-bstf stf)) 
			(marks (stfdat-staff stf))))

      (let ((current-clef (if (stfdat-tstf stf) :treble :bass))
	    (last-onset 0.0))
	(setf (staff-data (stfdat-staff stf)) (reverse (staff-data (stfdat-staff stf))))
	(let ((next-note (cdr (staff-data (stfdat-staff stf)))))
	  (loop for note in (staff-data (stfdat-staff stf)) do
	    ;;(print (format nil "pitch=~D" (+ (pitch note) (* 12 (octave note)))))
	    (if (and (stfdat-uclfs stf)
		     (audible-p note)
		     (/= last-onset (odb-onset note))
		     (not (and (stfdat-tstf stf) (stfdat-bstf stf))))
		(if (and (eq current-clef :bass)
			 (or (> (+ (pitch note) (* 12 (octave note))) 58)
			     (and (> (+ (pitch note) (* 12 (octave note))) 47)
				  next-note
				  (audible-p (first next-note))
				  (/= (onset note) (onset (first next-note)))
				  (> (+ (pitch (first next-note)) (* 12 (octave (first next-note)))) 48))))
		    (progn
		      ;;(print (format nil "push treble at ~1,3F with clef = ~A" (odb-onset note) current-clef))
		      (push (clef treble (onset (odb-onset note))) (staff-data (or (stfdat-tstf stf) (stfdat-bstf stf))))
		      (setf current-clef :treble))
		  (if (and (eq current-clef :treble)
			   (or (< (+ (pitch note) (* 12 (octave note))) 39)
			       (and (< (+ (pitch note) (* 12 (octave note))) 44)
				    next-note
				    (audible-p (first next-note))
				    (/= (onset note) (onset (first next-note)))
				    (< (+ (pitch (first next-note)) (* 12 (octave (first next-note)))) 48))))
		      (progn
			;;(print (format nil "push bass at ~1,3F with clef = ~A" (odb-onset note) current-clef))
			(push (clef bass (onset (odb-onset note))) (staff-data (or (stfdat-tstf stf) (stfdat-bstf stf))))
			(setf current-clef :bass)))))
	    (if (and (stfdat-tstf stf)
		     (or (not (audible-p note))
			 (and (> (+ (pitch note) (* 12 (octave note))) 47)
			      (stfdat-tstf stf))
			 (not (stfdat-bstf stf))))
		(push note (staff-data (stfdat-tstf stf)))
	      (push note (staff-data (stfdat-bstf stf))))
	    (if (audible-p note) (setf last-onset (odb-onset note))
	      (if (clef-p note) (setf current-clef (clef-name note))))
	    (setf next-note (cdr next-note))
	    ))
	)
      )
    
    (setf (systems score) (list (make-system :staves (nreverse new-staves))))
    (map-over-staves #'(lambda (score stf)
			 (declare (ignore score))
			 (setf (staff-data stf) (check-for-chords (staff-data stf))))
		     score)
    
    (if (not (member which-clefs '(:neither :none)))
	(loop for stf in staff-descriptors do
	  (if (stfdat-tstf stf) (push (clef (if (member which-clefs '(:treble :both)) treble
					      (if (eq which-clefs :bass) bass
						(if (eq which-clefs :alto) alto
						  tenor)))
					    (onset 0) (duration 0)) (staff-data (stfdat-tstf stf))))
	  (if (and (stfdat-bstf stf) 
		   (staff-data (stfdat-bstf stf)))
	      (push (clef bass (onset 0) (duration 0)) (staff-data (stfdat-bstf stf))))))

    (if (eq output-type :cmn)
	(let ((file (output-file score)))
	  (setf (output-file score) "hi.eps") ;a garbage file...
	  (cmn-store (cmn score)	;this is just to get the rests added explicitly
		     (if (string-equal (pathname-type file) "cmn") file "hi.cmn")
		     with-onsets))
      (cmn score))))

(defun make-name-a-string (name) (format nil "~S" name))

(defun display-notes (args notes)
  (let ((new-score (init-clm-input)))
    (loop for arg in args do
      (if (self-acting-p arg)
	  (funcall (action arg) new-score (arguments arg))))
    (loop for note in notes do (apply #'add-note new-score (make-name-a-string (first note)) (cdr note)))
    (finish-clm-input new-score nil nil)))

#|
;;; (defun set-staff-clef (staff main-clefs &rest other-clefs)
;;; (defun set-staff-number (staff number)
;;; (defun add-staff (score staff-name other-names &rest acts)
;;; (defun add-note (score name begin-time duration frequency &rest rest)
;;; (defun add-data (score name data)

;;; (dpyit (list (metronome 75)) (list (list 'hi (bar (onset 0))) (list 'hi (bass (onset 0))) (list 'hi (meter 3 4 (onset 0))) (list 'hi 0 .2 440 .1 staccato) '(silver .2 .6 660 .1) (list 'hi (bar (onset .8))) (list :score 'set-cmn-tempo .8 60) '(hi .8 1.0 440) '(ho 1.8 .5 220)))

(defun dpyit (args notes)
  (let ((new-score (init-clm-input)))
    (loop for arg in args do
      (if (self-acting-p arg)
	  (funcall (action arg) new-score (arguments arg))))
    (let ((hi-staff (add-staff new-score "HI" '("HO" "SILVER"))))
      (loop for note in notes do 
	(if (eq (first note) :score)
	    (apply (second note) *cmn-score* (cddr note))
	  (if (numberp (second note))
	      (apply #'add-note new-score (make-name-a-string (first note)) (cdr note))
	    (apply #'add-data new-score (make-name-a-string (first note)) (cdr note)))))
      (finish-clm-input new-score nil nil '(set-staff-clef "HI" :none)))))
|#

(defun add-clef (score staff type &optional size-factor)
  (declare (ignore score))
  (push (clef (if (clef-p type) type 
		(case type
		  (:bass bass)
		  (:treble treble)
		  (:alto alto)
		  (:tenor tenor)
		  (:percussion percussion)
		  (otherwise (error "unknown clef type: ~A" type))))
	      (if size-factor (scale size-factor size-factor)))
	(staff-data staff)))

(defun add-bar (score staff &optional (type :bar) bar-meter)
  (declare (ignore score))
  (push (bar (if (bar-p type) type
	       (case type
		 (:bar bar)
		 (:double-bar double-bar)
		 (:interior-double-bar interior-double-bar)
		 (:dashed-bar dashed-bar)
		 (:begin-repeat-bar begin-repeat-bar)
		 (:end-repeat-bar end-repeat-bar)
		 (:begin-and-end-repeat-bar begin-and-end-repeat-bar))))
	(staff-data staff))
  (if bar-meter
      (push (apply #'meter bar-meter) (staff-data staff))))

(defun set-tempo (score staff time tempo &optional pulse draw?)
  (let ((rhy (if (not pulse) q
	       (case pulse
		 (:q q)
		 (:e e)
		 (:q. q.)
		 (:h h)
		 (:w w)
		 (:h. h.)
		 (:s s)
		 (otherwise (error "unknown tempo pulse: ~A" pulse))))))
    (set-cmn-tempo score time (* tempo (quarters rhy)))
    (when draw?
      (if (not (staff-data staff))
	  (add-to-marks staff (list (mm tempo rhy)))
	(if (meter-p (first (staff-data staff)))
	    (add-to-marks (second (staff-data staff)) (list (mm tempo rhy)))
	  (add-to-marks (first (staff-data staff)) (list (mm tempo rhy))))))))
    
#|

(progn
  (setf hi (init-clm-input))
  (setf ho (add-staff hi "HIHO" nil))
  (add-clef hi ho :treble)
  (add-note-to-staff hi ho 0 1 440)
  (add-data hi "HIHO" begin-delayed-slur)
  (add-note-to-staff hi ho 0 1 660)
  (add-data hi "HIHO" end-delayed-slur)
  (add-note-to-staff hi ho 2 2 330)
  (finish-clm-input hi nil nil '(set-staff-clef "HIHO" :none))
  )
|#

#|
;;; bugs in this: 1) it changes the argument scores, causing subsequent calls to be messed up
;;;               2) it doesn't have the sense to omit staff and clef (and presumably meter/key) setup junk
(defun cmn-merge (&rest cmns)
  ;; take a list of cmn calls and combine them into one.  It is assumed
  ;; that all the calls have the same score setup (i.e. systems and staves are
  ;; compatible), and that the subsequent scores start at 0 but are intended to be
  ;; appended  to the end of the preceding score.
  (let ((new-score (first cmns))
	(durs nil))
    ;; first get the begin times fixed.
    (loop for cscore in cmns do
      (let ((cur-dur 0))
	(map-over-staves #'(lambda (score staff)
			     (let ((staff-dur (loop for n in (staff-data staff) if (audible-p n) maximize (+ (onset n) (duration n)))))
			       (setf cur-dur (max cur-dur staff-dur))))
			 cscore)
	(push cur-dur durs)))
    (let ((beg 0))
      (loop for cscore in (cdr cmns) do 
	(incf beg (pop durs))
	(map-over-staves #'(lambda (score staff)
			     (loop for n in (staff-data staff) do (if (onset n) (incf (onset n) beg))))
			 cscore)))
    ;; now do the two level merge (systems and staves)
    (loop for cscore in (cdr cmns) do
      (loop for sys in (systems new-score) and app-sys in (systems cscore) do
	(loop for stf in (staves sys) and app-stf in (staves app-sys) do
	  (setf (staff-data stf) (append (staff-data stf) (staff-data app-stf))))))
    new-score))
|#
      


;;;
;;; ----------------   cmn-store
;;;

;(defvar cmn-store-tags nil)		;a list of (obj str) pairs (in cmn1)
(defvar cmn-store-tag-ctr 0)
(defun new-cmn-store-tag (&optional (name "tag-")) (format nil "~A~D" name (incf cmn-store-tag-ctr)))

(defun check-for-cmn-store-tag (obj)	;returns a string suitable for format ~A
  (let ((pair (find obj cmn-store-tags :key #'first :test #'equal)))
    (if pair
	(progn
	  (setf cmn-store-tags (remove pair cmn-store-tags))
	  (format nil " ~A~A" 
		  (second pair)
		  (check-for-cmn-store-tag obj)))
      "")))

(defun add-to-cmn-store-tags (obj str)
  (push (list obj str) cmn-store-tags))

(defun cmn-store (&optional (scr *cmn-score*) (file "hi.cmn") (with-onsets t))
  (setf cmn-store-tags nil)
  (setf cmn-store-tag-ctr 0)
  (with-open-file (s file :direction :output :if-exists :supersede :if-does-not-exist :create)
    (format s ";;; cmn structure stored ~A~%(cmn" (creation-date))
    (if (title scr) 
	(if (not (text-p (title scr)))
	    (format s " (title ~S)" (title scr))
	  (format s " (title ~A)" (identify (title scr)))))
    (format s "~{~A~}" (style-list scr nil t))
    (identify-marks scr s)
    (let ((tied-to-staves (remove-duplicates 
			   (loop for system in (systems scr) append
			     (loop for staff in (staves system)
			      if (true-staff staff) collect (true-staff staff))))))
      (loop for system in (systems scr) do
	(format s "~%  (system")
	(identify-marks system s)
	(loop for staff in (staves system) do
	  (let ((staff-y0 (staff-y0 staff))
		(last-object nil)
		(possible-line-break nil)
		(first-object t))
	    (if (not (member staff tied-to-staves))
		(format s "~%    (staff")
	      (format s " (setf staff-~D (staff" (position staff tied-to-staves)))
	    (if (true-staff staff)
		(format s " (tied-to staff-~D)" (position (true-staff staff) tied-to-staves))
	      (progn
		(format s "~A" (identify-dxy staff))
		(if (staff-size staff) (format s " (staff-size ~1,3F)" (staff-size staff)))
		(if (staff-lines staff) (format s " (staff-lines ~D)" (staff-lines staff)))
		(if (staff-name staff) 
		    (if (not (text-p (staff-name staff)))
			(format s " (staff-name ~S)" (staff-name staff))
		      (format s " (staff-name ~A)" (identify (staff-name staff)))))))
	    (identify-marks staff s)
	    (loop for object in (staff-data staff) do
	      (let ((cur-staff-y0 (or (staff-y0 object) (box-y0 object))))
		(if (and (not (staff-p object)) ;might want to put line-mark here
			 (or (not (bar-p object))
			     (not (bar-p last-object))) ;there's an extra left-justified bar sometimes
			 (or first-object
			     (= cur-staff-y0 staff-y0)
			     (not (or (clef-p object)
				      (key-p object)
				      (brace-p object)
				      (bracket-p object)
				      (bar-p object)))))
		    (let ((cur-onset (odb-onset object)))
		      (setf first-object nil)
		      (setf staff-y0 cur-staff-y0)
		      (setf last-object object)
		      (if (not with-onsets) (setf (odb-onset object) nil))
		      (format s "~%      ~A" (identify object))
		      (if (not with-onsets) (setf (odb-onset object) cur-onset))
		      (if (audible-p object)
			  (setf possible-line-break (and (store-data object)
							 (member :line-break (store-data object)))))
		      ))
		(if (staff-p object)
		    (progn
		      (if possible-line-break
			  (format s "~%      (line-mark)"))
		      (setf possible-line-break nil)))
		))
	    (if (not (member staff tied-to-staves))
		(format s ")")
	      (format s "))"))))
	(format s ")"))
      (format s ")~%"))))
